package org.jlab.coda.jevio;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * This the header for an evio bank structure (<code>EvioBank</code>). It does not contain the raw data, just the
 * header. Note: since an "event" is really just the outermost bank, this is also the header for an
 * <code>EvioEvent</code>.
 * 
 * @author heddle
 * 
 */
public final class BankHeader extends BaseStructureHeader {
	
	/**
	 * Null constructor.
	 */
	public BankHeader() {
	}

	/**
	 * Constructor
	 * @param tag the tag for the bank header.
	 * @param dataType the data type for the content of the bank.
	 * @param num sometimes, but not necessarily, an ordinal enumeration.
	 */
	public BankHeader(int tag, DataType dataType, int num) {
		super(tag, dataType, num);
	}

	/**
	 * Constructor
	 * @param tag the tag for the bank header.
	 * @param dataType the (int) data type for the content of the bank.
	 * @param num sometimes, but not necessarily, an ordinal enumeration.
	 */
	public BankHeader(int tag, int dataType, int num) {
		super(tag, dataType, num);
	}

    /**
     * {@inheritDoc}
     */
    public int getHeaderLength() {return 2;}

    /**
     * {@inheritDoc}
     */
    protected void toArray(byte[] bArray, int offset, ByteOrder order) {
        if (order == ByteOrder.BIG_ENDIAN) {
            ByteDataTransformer.intToBytes(length, bArray, offset);
            ByteDataTransformer.shortToBytes(shortValue(tag), bArray, offset+4);
            bArray[offset+6] = byteValue(dataType);
            bArray[offset+7] = byteValue(number);
        }
        else {
            ByteDataTransformer.intToBytes(Integer.reverseBytes(length), bArray, offset);
            bArray[offset+4] = byteValue(number);
            bArray[offset+5] = byteValue(dataType);
            ByteDataTransformer.shortToBytes(Short.reverseBytes(shortValue(tag)), bArray, offset+6);
        }
        
    }


	/**
	 * Obtain a string representation of the bank header.
	 * 
	 * @return a string representation of the bank header.
	 */
	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer(512);
		sb.append(String.format("bank length: %d\n", length));
		sb.append(String.format("number:      %d\n", number));
		sb.append(String.format("data type:   %s\n", getDataTypeName()));
		sb.append(String.format("tag:         %d\n", tag));
		return sb.toString();
	}
	
	/**
	 * Write myself out a byte buffer. This write is relative--i.e., it uses the current position of the buffer.
	 * 
	 * @param byteBuffer the byteBuffer to write to.
	 * @return the number of bytes written, which for a BankHeader is 8.
	 */
	@Override
	public int write(ByteBuffer byteBuffer) {
		byteBuffer.putInt(length);
        
        if (byteBuffer.order() == ByteOrder.BIG_ENDIAN) {
            byteBuffer.putShort(shortValue(tag));
            byteBuffer.put(byteValue(dataType));
            byteBuffer.put(byteValue(number));
        }
        else {
            byteBuffer.put(byteValue(number));
            byteBuffer.put(byteValue(dataType));
            byteBuffer.putShort(shortValue(tag));
        }

		return 8;
	}

}
