package org.jlab.coda.jevio;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Vector;
import java.util.LinkedList;
import java.io.UnsupportedEncodingException;
import java.io.StringWriter;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.stream.XMLOutputFactory;

/**
 * This is the base class for all evio structures: Banks, Segments, and TagSegments. It implements
 * <code>MutableTreeNode</code> because a tree representation of events is created when a new event is parsed.
 * 
 * @author heddle
 * @author timmer - add byte order tracking
 *
 */
public abstract class BaseStructure implements IEvioStructure, MutableTreeNode, IEvioWriter {

	/**
	 * Holds the header of the bank.
	 */
	protected BaseStructureHeader header;

	/**
	 * The raw data of the structure.
	 */
	protected byte rawBytes[];

	/**
	 * Used if raw data should be interpreted as ints.
	 */
	protected int intData[];

	/**
	 * Used if raw data should be interpreted as longs.
	 */
	protected long longData[];

	/**
	 * Used if raw data should be interpreted as shorts.
	 */
	protected short shortData[];

	/**
	 * Used if raw data should be interpreted as doubles.
	 */
	protected double doubleData[];

    /**
     * Used if raw data should be interpreted as doubles.
     */
    protected float floatData[];

    /**
     * Used if raw data should be interpreted as a string.
     */
    protected StringBuffer stringData;

    /**
     * Keep track of end of the last string added to stringData (including null but not padding).
     */
    private int stringEnd;

    /**
     * The number of stored data items like number of banks, ints, floats, etc.
     * (not the size in ints or bytes). Some items may be padded such as shorts
     * and bytes. This will tell the meaningful number of such data items.
     * In the case of containers, returns number of bytes not in header.
     */
    protected int numberDataItems;

    /**
     * Give the XML output proper indentation.
     */
    protected String xmlIndent = "";

    /**
     * Name of this object as an XML element.
     */
    protected String xmlElementName;

    /**
     * Name of this object's contents as an XML attribute if it is a structure type.
     */
    protected String xmlContentAttributeName;

    /**
     * Endianness of the data if appropriate,
     * either {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
     */
    protected ByteOrder byteOrder;

	/**
	 * The parent of the structure. If it is an "event", the parent is null. This is used for creating trees for a
	 * variety of purposes (not necessarily graphical.)
	 */
	private BaseStructure parent;

	/**
	 * Holds the children of this structure. This is used for creating trees for a variety of purposes
     * (not necessarily graphical.)
	 */
	protected Vector<BaseStructure> children;

    /**
	 * Constructor using a provided header
	 * 
	 * @param header the header to use.
	 * @see BaseStructureHeader
	 */
	public BaseStructure(BaseStructureHeader header) {
		this.header = header;
        // by default we're big endian since this is java
        byteOrder = ByteOrder.BIG_ENDIAN;
        setXmlNames();
	}

	/**
	 * A convenience method use instead of "instanceof" to see what type of structure we have. Note: this returns the
	 * type of this structure, not the type of data (the content type) this structure holds.
	 * 
	 * @return the <code>StructureType</code> of this structure.
	 * @see StructureType
	 */
	@Override
	public abstract StructureType getStructureType();

    /**
     * This is a method that must be filled in by subclasses. Write this structure out as an XML record.
     *
     * @param xmlWriter the writer used to write the events. It is tied to an open file.
     */
    public abstract void toXML(XMLStreamWriter xmlWriter);

	/**
	 * Get the element name for the structure for writing to XML.
	 * 
	 * @return the element name for the structure for writing to XML.
	 */
	public abstract String getXMLElementName();

    /**
     * Write this structure out as an XML format string.
     */
    public String toXML() {

        try {
            StringWriter sWriter = new StringWriter();
            XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance().createXMLStreamWriter(sWriter);

            toXML(xmlWriter);
            return sWriter.toString();
        }
        catch (XMLStreamException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Set the name of this object's XML element and also content type attribute if content is structure type.
     */
    protected void setXmlNames() {

        // is this structure holding a structure type container?
        boolean holdingStructureType = false;

        // type of data held by this container type
        DataType type = header.getDataTypeEnum();

        switch (type) {
            case CHAR8:
                xmlElementName = "int8"; break;
            case UCHAR8:
                xmlElementName = "uint8"; break;
            case SHORT16:
                xmlElementName = "int16"; break;
            case USHORT16:
                xmlElementName = "uint16"; break;
            case INT32:
                xmlElementName = "int32"; break;
            case UINT32:
                xmlElementName = "uint32"; break;
            case LONG64:
                xmlElementName = "int64"; break;
            case ULONG64:
                xmlElementName = "uint64"; break;
            case FLOAT32:
                xmlElementName = "float32"; break;
            case DOUBLE64:
                xmlElementName = "float64"; break;
            case CHARSTAR8:
                xmlElementName = "string"; break;

            case UNKNOWN32:
                holdingStructureType = true;
                xmlContentAttributeName = "unknown32";
                break;

            case TAGSEGMENT:
            case ALSOTAGSEGMENT:
                holdingStructureType = true;
                xmlContentAttributeName = "tagsegment";
                break;

            case SEGMENT:
            case ALSOSEGMENT:
                holdingStructureType = true;
                xmlContentAttributeName = "segment";
                break;

            case BANK:
            case ALSOBANK:
                holdingStructureType = true;
                xmlContentAttributeName = "bank";
                break;

            default:
                xmlElementName = "unknown"; break;
        }

        if (holdingStructureType) {
            // Which container type are we (NOT what are we holding)?
            // Because that determines our element name.
            StructureType structType = getStructureType();
            if (structType == StructureType.UNKNOWN32) {
                xmlElementName = "unknown32";
            }
            else {
                xmlElementName = getXMLElementName();
            }
        }

    }

    /**
     * Set the indentation (string of spaces) for more pleasing XML output.
     * @param s the indentation (string of spaces) for more pleasing XML output
     */
    protected void setXmlIndent(String s) {
        xmlIndent = s;
    }

    /**
     * Increase the indentation (string of spaces) of the XML output.
     */
    protected void increaseXmlIndent() {
        xmlIndent += "   ";
    }

    /**
     * Decrease the indentation (string of spaces) of the XML output.
     */
    protected void decreaseXmlIndent() {
        xmlIndent = xmlIndent.substring(0, xmlIndent.length() - 3);
    }

    /**
     * What is the byte order of this data?
     * @return {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}
     */
    public ByteOrder getByteOrder() {
        return byteOrder;
    }

    /**
     * Set the byte order of this data. This method <b>cannot</b> be used to swap data.
     * It is only used to describe the endianness of the rawdata contained.
     * @param byteOrder {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}
     */
    public void setByteOrder(ByteOrder byteOrder) {
        this.byteOrder = byteOrder;
    }

	/**
	 * Is a byte swap required? This is java and therefore big endian. If data is
     * little endian, then a swap is required.
	 * 
	 * @return <code>true</code> if byte swapping is required (data is little endian).
	 */
	public boolean isSwap() {
		return byteOrder == ByteOrder.LITTLE_ENDIAN;
	}

	/**
	 * Get the description from the name provider (dictionary), if there is one.
	 * 
	 * @return the description from the name provider (dictionary), if there is one. If not, return
	 *         NameProvider.NO_NAME_STRING.
	 */
	@Override
	public String getDescription() {
		return NameProvider.getName(this);
	}

	/**
	 * Obtain a string representation of the structure.
	 * 
	 * @return a string representation of the structure.
	 */
	@Override
	public String toString() {
		StructureType stype = getStructureType();
		DataType dtype = header.getDataTypeEnum();

		String description = getDescription();
		if (NameProvider.NO_NAME_STRING.equals(description)) {
			description = null;
		}
		String s = stype + " of " + dtype + "s " + " len (ints): " + header.length + " " + " tag: " + header.tag;
		if (this instanceof EvioBank) {
			s += " num: " + header.number;
		}

		if (rawBytes == null) {
			s += " data: null";
		}
		else {
			s += " datalen (bytes): " + rawBytes.length;
		}

		if (description != null) {
			s += " [" + description + "]";
		}

		int numChildren = (children == null) ? 0 : children.size();

		s += " <#children: " + numChildren + ">";
		return s;
	}

	/**
	 * This is a method from the IEvioStructure Interface. Return the header for this structure.
	 * 
	 * @return the header for this structure.
	 */
	@Override
	public BaseStructureHeader getHeader() {
		return header;
	}

    /**
     * Get the number of stored data items like number of banks, ints, floats, etc.
     * (not the size in ints or bytes). Some items may be padded such as shorts
     * and bytes. This will tell the meaningful number of such data items.
     * In the case of containers, returns number of bytes not in header.
     *
     * @return number of stored data items (not size or length),
     *         or number of bytes if container
     */
    public int getNumberDataItems() {
        if (isContainer()) {
            numberDataItems = header.getLength() + 1 - header.getHeaderLength();
        }

        // if the calculation has not already been done ...
        if (numberDataItems < 1) {
            // When parsing a file or byte array, it is not fully unpacked until data
            // is asked for specifically, for example as an int array or a float array.
            // Thus we don't know how many of a certain item (say doubles) there is.
            // But we can figure that out now based on the size of the raw data byte array.
            int divisor = 0;
            DataType type = header.getDataTypeEnum();

            switch (type) {
                case CHAR8:
                case UCHAR8:
                    divisor = 1; break;
                case SHORT16:
                case USHORT16:
                    divisor = 2; break;
                case INT32:
                case UINT32:
                case FLOAT32:
                    divisor = 4; break;
                case LONG64:
                case ULONG64:
                case DOUBLE64:
                    divisor = 8; break;
                case CHARSTAR8:
                    String[] s = getStringData();
                    numberDataItems = s.length;
                    break;
                default:
            }

            if (divisor > 0) numberDataItems = rawBytes.length/divisor;
        }

        return numberDataItems;
    }

    /**
     * Get the length of this structure in bytes, including the header.
     * @return the length of this structure in bytes, including the header.
     */
    public int getTotalBytes() {
        return 4*(header.getLength() + 1);
    }

	/**
	 * Get the raw data of the structure.
	 * 
	 * @return the raw data of the structure.
	 */
	public byte[] getRawBytes() {
		return rawBytes;
	}

	/**
	 * Set the data for the structure.
	 * 
	 * @param rawBytes the structure raw data.
	 */
	public void setRawBytes(byte[] rawBytes) {
		this.rawBytes = rawBytes;
	}

	/**
	 * This is a method from the IEvioStructure Interface. Gets the raw data as an integer array,
     * if the content type as indicated by the header is appropriate.
     * NOTE: since Java does not have unsigned primitives, both INT32 and UINT32
	 * data types will be returned as int arrays. The application will have to deal
     * with reinterpreting signed ints that are negative as unsigned ints.
	 * 
	 * @return the data as an int array, or <code>null</code> if this makes no sense for the given content type.
	 */
	@Override
	public int[] getIntData() {

		switch (header.getDataTypeEnum()) {
		case INT32:
		case UINT32:
			if (intData == null) {
				if (rawBytes == null) {
					return null;
				}
				intData = ByteDataTransformer.getAsIntArray(rawBytes, byteOrder);
			}
			return intData;
		default:
			return null;
		}
	}

	/**
	 * This is a method from the IEvioStructure Interface. Gets the raw data as a long array,
     * if the content type as indicated by the header is appropriate.
     * NOTE: since Java does not have unsigned primitives, both LONG64 and ULONG64
     * data types will be returned as long arrays. The application will have to deal
     * with reinterpreting signed longs that are negative as unsigned longs.
	 * 
	 * @return the data as an long array, or <code>null</code> if this makes no sense for the given content type.
	 */
	@Override
	public long[] getLongData() {

		switch (header.getDataTypeEnum()) {
		case LONG64:
		case ULONG64:
			if (longData == null) {
				if (rawBytes == null) {
					return null;
				}
				longData = ByteDataTransformer.getAsLongArray(rawBytes, byteOrder);
			}
			return longData;
		default:
			return null;
		}
	}

	/**
	 * This is a method from the IEvioStructure Interface. Gets the raw data as a float array,
     * if the content type as indicated by the header is appropriate.
	 * 
	 * @return the data as an double array, or <code>null</code> if this makes no sense for the given contents type.
	 */
	@Override
	public float[] getFloatData() {

		switch (header.getDataTypeEnum()) {
		case FLOAT32:
			if (floatData == null) {
				if (rawBytes == null) {
					return null;
				}
				floatData = ByteDataTransformer.getAsFloatArray(rawBytes, byteOrder);
			}
			return floatData;
		default:
			return null;
		}
	}

	/**
	 * This is a method from the IEvioStructure Interface. Gets the raw data as a double array,
     * if the content type as indicated by the header is appropriate.
	 * 
	 * @return the data as an double array, or <code>null</code> if this makes no sense for the given content type.
	 */
	@Override
	public double[] getDoubleData() {

		switch (header.getDataTypeEnum()) {
		case DOUBLE64:
			if (doubleData == null) {
				if (rawBytes == null) {
					return null;
				}
				doubleData = ByteDataTransformer.getAsDoubleArray(rawBytes, byteOrder);
			}
			return doubleData;
		default:
			return null;
		}
	}

	/**
	 * This is a method from the IEvioStructure Interface. Gets the raw data as a double array,
     * if the contents type as indicated by the header is appropriate.
     * NOTE: since Java does not have unsigned primitives, both SHORT16 and USHORT16
     * data types will be returned as short arrays. The application will have to deal
     * with reinterpreting signed shorts that are negative as unsigned longs.
	 * 
	 * @return the data as an short array, or <code>null</code> if this makes no sense for the given contents type.
	 */
	@Override
	public short[] getShortData() {

		switch (header.getDataTypeEnum()) {
		case SHORT16:
		case USHORT16:
			if (shortData == null) {
				if (rawBytes == null) {
					return null;
				}
				shortData = ByteDataTransformer.getAsShortArray(rawBytes, byteOrder);
			}
			return shortData;
		default:
			return null;
		}
	}

	/**
	 * This is a method from the IEvioStructure Interface. Gets the raw data as an byte array,
     * if the contents type as indicated by the header is appropriate.
     * NOTE: since Java does not have unsigned primitives, CHARSTAR8, CHAR8 and UCHAR8
	 * data types will be returned as byte arrays. The application will have to deal
     * with reinterpreting bytes as characters, if necessary.
	 * 
	 * @return the data as an byte array, or <code>null</code> if this makes no sense for the given contents type.
	 */
	@Override
	public byte[] getByteData() {

		switch (header.getDataTypeEnum()) {
		case CHARSTAR8:
		case CHAR8:
		case UCHAR8:
			return rawBytes;
		default:
			return null;
		}
	}

    /**
     * This is a method from the IEvioStructure Interface. Gets the raw data (ascii) as an
     * array of String objects, if the contents type as indicated by the header is appropriate.
     * For any other behavior, the user should retrieve the data as a byte array and
     * manipulate it in the exact manner desired.<p>
     * Originally, in evio versions 1, 2 and 3, only one string was stored. Recent changes cause
     * an array of strings to be stored and retrieved. The changes are backwards compatible.
     *
     * The following is true about the string raw data format:
     * <ul>
     * <li>All strings are immediately followed by an ending null (0).
     * <li>All string arrays are further padded/ended with at least one 0x4 valued ASCII
     *     char (up to 4 possible).
     * <li>The presence of 1 to 4 ending 4's distinguishes the recent string array version from
     *     the original, single string version.
     * <li>The priginal version string may be padded with anything after its ending null.
     * <li>Ending non-printing ascii chars (value < 32, = 127) are not included in the string
     *     since they are there for padding.
     * </ul>
     *
     * @return the data as an array of String objects if DataType is CHARSTAR8, or <code>null</code>
     *         if this makes no sense for the given type.
     *
     */
    @Override
    public String[] getStringData() {

        String[] strs;
        switch (header.getDataTypeEnum()) {
        case CHARSTAR8:
            if (stringData == null) {
                if (rawBytes == null) {
                    return null;
                }
                try {
                    // stringData contains all elements of rawBytes
                    stringData = new StringBuffer(new String(rawBytes, "US-ASCII"));
                }
                catch (UnsupportedEncodingException e) { /* will never happen */ }
            }

            // Strings are separated by ASCII nulls (char val = 0) and the end is padded by ASCII 4's
            // (char val = 4). However, in the legacy versions of evio, there is
            // only 1 null-terminated string and anything as padding. To accomodate legacy evio, if
            // there is not an ending ASCII value 4, anything past the first null is ignored.
            // After doing so, split at the nulls. Do not use the String
            // method "split" as any empty trailing strings are unfortunately discarded.

            boolean newVersion = true;
            if (stringData.charAt(stringData.length() - 1) != '\004') {
                newVersion = false;
            }

            char c;
            LinkedList<Integer> nullIndexList = new LinkedList<Integer>();

            for (int i=0; i < stringData.length(); i++) {
                c = stringData.charAt(i);
                if ( c == '\000' ) {
                    nullIndexList.add(i);
                    // only 1 null terminated string originally in evio v2,3
                    if (!newVersion) {
                        break;
                    }
                }
                // Look for any non-printing/control characters (not including null)
                // and end the string there. Allow new lines (c = 10).
                else if ( (c < 32 || c == 127) && c != 10 ) {
                    break;
                }
            }

            strs = new String[nullIndexList.size()];
            int lastIndex, firstIndex=0;
            for (int i=0; i < nullIndexList.size(); i++) {
                lastIndex  = nullIndexList.get(i);
                strs[i]    = stringData.substring(firstIndex, lastIndex);
                firstIndex = lastIndex + 1;
            }

            return strs;

        default:
            return null;
        }
    }

    /**
     * This is a method from the IEvioStructure Interface. Gets the raw data (ascii) as a String,
     * if the contents type as indicated by the header is appropriate.
     * NOTE: ending non-printing ascii chars (value < 32) are not included in the string
     * (since they are most likely there for padding).
     * For any other behavior, the user should retrieve the data as a byte array and
     * manipulate it in the exact manner desired.
     *
     * @return the data as a String if DataType is CHARSTAR8, or <code>null</code>
     *         if this makes no sense for the given type.
     *
     */
    //@Override
    private String getStringDataOrig() {

        switch (header.getDataTypeEnum()) {
        case CHARSTAR8:
            if (stringData == null) {
                if (rawBytes == null) {
                    return null;
                }
                try {
                    // stringData contains all elements of rawBytes
                    stringData = new StringBuffer(new String(rawBytes, "US-ASCII"));
                }
                catch (UnsupportedEncodingException e) { /* will never happen */ }
            }

            // When fashioning a return string, we assume ascii (one char per byte).
            // If there are any bytes = 0 (ie null termination) or other nonprinting chars,
            // Java includes them in the string, but ignores them when printing.
            // In light of this and the fact that this data type is padded to make an
            // integral number of 32 byte chunks, the ending nonprinting chars will be
            // stripped off when forming the return string.
            String s = null;
            try {
                int endingNonPrintingChars = 0;
                for (int i=rawBytes.length-1; i >= 0; i--) {
                    if (rawBytes[i] < 32) {
                        endingNonPrintingChars++;
                        continue;
                    }
                    break;
                }
                s = new String(rawBytes, 0, rawBytes.length - endingNonPrintingChars, "US-ASCII");
            }
            catch (UnsupportedEncodingException e) { /* will never happen */ }

            return s;

        default:
            return null;
        }
    }

	/**
	 * Get the parent of this structure. The outer event bank is the only structure with a <code>null</code>
     * parent (the only orphan). All other structures have non-null parent giving the container in which they
     * were embedded. Part of the <code>MutableTreeNode</code> interface.
	 * 
	 * @return the parent of this structure.
	 */
	@Override
	public BaseStructure getParent() {
		return parent;
	}

	/**
	 * Get an enumeration of all the children of this structure. If none, it returns a constant,
     * empty Enumeration. Part of the <code>MutableTreeNode</code> interface.
	 */
	@Override
	public Enumeration<?> children() {
		if (children == null) {
			return DefaultMutableTreeNode.EMPTY_ENUMERATION;
		}
		return children.elements();
	}

	/**
	 * Add a child at the given index.
	 * 
	 * @param child the child to add.
	 * @param index the target index. Part of the <code>MutableTreeNode</code> interface.
	 */
	@Override
	public void insert(MutableTreeNode child, int index) {
		if (children == null) {
			children = new Vector<BaseStructure>(10);
		}
		children.add(index, (BaseStructure) child);
	}

	/**
	 * Convenience method to add a child at the end of the child list. In this application
	 * where are not concerned with the ordering of the children.
	 * 
	 * @param child the child to add. It will be added to the end of child list.
	 */
	public void insert(MutableTreeNode child) {
		if (children == null) {
			children = new Vector<BaseStructure>(10);
		}
		//add to end
		insert(child, children.size());
	}

	/**
	 * Removes the child at index from the receiver. Part of the <code>MutableTreeNode</code> interface.
	 * 
	 * @param index the target index for removal.
	 */
	@Override
	public void remove(int index) {
		children.remove(index);
	}

	/**
	 * Removes the child. Part of the <code>MutableTreeNode</code> interface.
	 * 
	 * @param child the child node being removed.
	 */
	@Override
	public void remove(MutableTreeNode child) {
		children.remove(child);
	}

	/**
	 * Remove this node from its parent. Part of the <code>MutableTreeNode</code> interface.
	 */
	@Override
	public void removeFromParent() {
		parent.remove(this);
	}

	/**
	 * Set the parent for this node. Part of the <code>MutableTreeNode</code> interface.
	 */
	@Override
	public void setParent(MutableTreeNode parent) {
		this.parent = (BaseStructure) parent;
	}

	/**
	 * This method is not relevant for this implementation.
     * An empty implementation is provided to satisfy the interface.
	 * Part of the <code>MutableTreeNode</code> interface.
	 */
	@Override
	public void setUserObject(Object arg0) {
	}

	/**
	 * Checks whether children are allowed. Structures of structures can have children.
     * Structures of primitive data can not. Thus is is entirely determined by the content type.
     * Part of the <code>MutableTreeNode</code> interface.
	 * 
	 * @return <code>true</code> if this node does not hold primitive data,
     *         i.e., if it is a structure of structures (a container).
	 */
	@Override
	public boolean getAllowsChildren() {
		return header.getDataTypeEnum().isStructure();
	}

	/**
	 * Obtain the child at the given index. Part of the <code>MutableTreeNode</code> interface.
	 * 
	 * @param index the target index.
	 * @return the child at the given index.
     * @throws ArrayIndexOutOfBoundsException if the index is out of range
     *	       ({@code index < 0 || index >= children.size()})
	 */
	@Override
	public TreeNode getChildAt(int index) {
		return children.elementAt(index);
	}

	/**
	 * Get the count of the number of children. Part of the <code>MutableTreeNode</code> interface.
	 * 
	 * @return the number of children.
	 */
	@Override
	public int getChildCount() {
		if (children == null) {
			return 0;
		}
		return children.size();
	}

	/**
	 * Get the index of a node. Part of the <code>MutableTreeNode</code> interface.
	 * 
	 * @return the index of the target node.
	 */
	@Override
	public int getIndex(TreeNode node) {
		return children.indexOf(node);
	}

	/**
	 * Checks whether this is a leaf. Leaves are structures with no children.
     * All structures that contain primitive data are leaf structures.
	 * Part of the <code>MutableTreeNode</code> interface.<br>
	 * Note: this means that an empty container, say a Bank of Segments that have no segments, is a leaf.
	 * 
	 * @return <code>true</code> if this is a structure with a primitive data type, i.e., it is not a container
	 *         structure that contains other structures.
	 */
	@Override
	public boolean isLeaf() {
		return ((children == null) || (children.size() < 1));
	}
	
	/**
	 * Checks whether this structure is a container, i.e. a structure of structures.
	 * @return <code>true</code> if this structure is a container. This is the same check as 
	 * <code>getAllowsChildren()</code>.
	 */
	public boolean isContainer() {
		return header.getDataTypeEnum().isStructure();
	}

	/**
	 * This calls the xml write for all of a structure's children.
     * It is used for a depth-first traversal through the tree
	 * when writing an event to xml.
	 * 
	 * @param xmlWriter the writer used to write the events. It is tied to an open file.
	 */
	protected void childrenToXML(XMLStreamWriter xmlWriter) {
		if (children == null) {
			return;
		}
        
        increaseXmlIndent();

		for (BaseStructure structure : children) {
			structure.setXmlIndent(xmlIndent);
            structure.toXML(xmlWriter);
		}

        decreaseXmlIndent();
	}

	/**
	 * All structures have a common start to their xml writing
	 * 
	 * @param xmlWriter the writer used to write the events. It is tied to an open file.
	 */
	protected void commonXMLStart(XMLStreamWriter xmlWriter) {
		try {
			xmlWriter.writeCharacters("\n");
            xmlWriter.writeCharacters(xmlIndent);
			xmlWriter.writeStartElement(xmlElementName);
		}
		catch (XMLStreamException e) {
			e.printStackTrace();
		}
	}

	/**
	 * All structures have a common data write for their xml writing
	 * 
	 * @param xmlWriter the writer used to write the events. It is tied to an open file.
	 */
	protected void commonXMLDataWrite(XMLStreamWriter xmlWriter) {

		// only leafs write data
		if (!isLeaf()) {
			return;
		}

		try {
            String s;
            String indent = String.format("\n%s", xmlIndent);

			BaseStructureHeader header = getHeader();
			switch (header.getDataTypeEnum()) {
			case DOUBLE64:
				double doubledata[] = getDoubleData();
                for (int i=0; i < doubledata.length; i++) {
                    if (i%2 == 0) {
                        xmlWriter.writeCharacters(indent);
                    }
                    s = String.format("%24.16e  ", doubledata[i]);
                    xmlWriter.writeCharacters(s);
                }
				break;

			case FLOAT32:
				float floatdata[] = getFloatData();
                for (int i=0; i < floatdata.length; i++) {
                    if (i%2 == 0) {
                        xmlWriter.writeCharacters(indent);
                    }
                    s = String.format("%14.7e  ", floatdata[i]);
                    xmlWriter.writeCharacters(s);
                }
				break;

			case LONG64:
			case ULONG64:
				long longdata[] = getLongData();
                for (int i=0; i < longdata.length; i++) {
                    if (i%2 == 0) {
                        xmlWriter.writeCharacters(indent);
                    }
                    s = String.format("%20d  ", longdata[i]);
                    xmlWriter.writeCharacters(s);
                }
				break;

			case INT32:
			case UINT32:
				int intdata[] = getIntData();
                for (int i=0; i < intdata.length; i++) {
                    if (i%5 == 0) {
                        xmlWriter.writeCharacters(indent);
                    }
                    s = String.format("%11d  ", intdata[i]);
                    xmlWriter.writeCharacters(s);
                }
				break;

			case SHORT16:
			case USHORT16:
				short shortdata[] = getShortData();
                for (int i=0; i < shortdata.length; i++) {
                    if (i%5 == 0) {
                        xmlWriter.writeCharacters(indent);
                    }
                    s = String.format("%6d  ", shortdata[i]);
                    xmlWriter.writeCharacters(s);
                }
				break;

			case CHAR8:
			case UCHAR8:
				byte bytedata[] = getByteData();
                for (int i=0; i < bytedata.length; i++) {
                    if (i%5 == 0) {
                        xmlWriter.writeCharacters(indent);
                    }
                    s = String.format("%4d  ", bytedata[i]);
                    xmlWriter.writeCharacters(s);
                }
				break;

			case CHARSTAR8:
                String stringdata[] = getStringData();
                for (String str : stringdata) {
                    xmlWriter.writeCharacters("\n");
                    xmlWriter.writeCData(str);
                }
				break;

			}
		}
		catch (XMLStreamException e) {
			e.printStackTrace();
		}

	}

	/**
	 * Compute the dataLength. This is the amount of data needed by a leaf of primitives. For non-leafs (structures of
	 * structures) this returns 0. For data types smaller than an int, e.g. a short, it computes assuming padding to an
	 * integer number of ints. For example, if we are writing a byte array of length 3 or 4, the it would return 1. If
	 * the byte array is 5,6,7 or 8 it would return 2;
	 * 
	 * @return
	 */
	protected int dataLength() {

		int datalen = 0;

		// only leafs write data
		if (isLeaf()) {
			BaseStructureHeader header = getHeader();
			switch (header.getDataTypeEnum()) {
			case DOUBLE64:
				double doubledata[] = getDoubleData();
				if (doubledata != null) {
					datalen = 2 * doubledata.length;
				}
				break;

			case FLOAT32:
				float floatdata[] = getFloatData();
				if (floatdata != null) {
					datalen = floatdata.length;
				}
				break;

			case LONG64:
			case ULONG64:
				long longdata[] = getLongData();
				if (longdata != null) {
					datalen = 2 * longdata.length;
				}
				break;

			case INT32:
			case UINT32:
				int intdata[] = getIntData();
				if (intdata != null) {
					datalen = intdata.length;
				}
				break;

			case SHORT16:
			case USHORT16:
				short shortdata[] = getShortData();
				if (shortData != null) {
					datalen = 1 + (shortdata.length - 1) / 2;
				}
				break;

            case CHAR8:
            case UCHAR8:
            case CHARSTAR8: // rawbytes contain ascii, already padded
                byte bytedata[] = getByteData();
                if (bytedata != null) {
                    datalen = 1 + (bytedata.length - 1) / 4;
                }
                break;

			} // switch
		} // isleaf
		return datalen;

	}

	/**
	 * All structures close their xml writing the same way, by drilling down to embedded children.
	 * 
	 * @param xmlWriter the writer used to write the events. It is tied to an open file.
	 */
	protected void commonXMLClose(XMLStreamWriter xmlWriter) {
		childrenToXML(xmlWriter);
		try {
			xmlWriter.writeCharacters("\n");
            xmlWriter.writeCharacters(xmlIndent);
			xmlWriter.writeEndElement();
		}
		catch (XMLStreamException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Get the children of this structure.
	 * 
	 * @return the children of this structure.
	 */
	public Vector<BaseStructure> getChildren() {
		return children;
	}

	/**
	 * Compute the and set length of all header fields for this structure and all its descendants.
     * For writing events, this will be crucial for setting the values in the headers.
	 * 
	 * @return the length that would go in the header field (for a leaf).
	 */
	public int setAllHeaderLengths() {
		int datalen;

		if (isLeaf()) {
			datalen = dataLength();
		}
		else {
			datalen = 0;

			if (children == null) {
				System.err.println("Non leaf with null children!");
				System.exit(1);
			}
			for (BaseStructure child : children) {
				datalen += child.setAllHeaderLengths();
				datalen++; // for the length word of each child
			}
		}

        if (this instanceof EvioBank) {
            datalen += 1; // extra header word
        }

		// set the datalen for the header
		header.setLength(datalen);
		return datalen;
	}

	/**
	 * Write myself out a byte buffer.
     * TODO: This method can probably be made more efficient
	 * 
	 * @param byteBuffer the byteBuffer to write to.
	 * @return the number of bytes written.
	 */
	public int write(ByteBuffer byteBuffer) {

		int startPos = byteBuffer.position();

		// write the header
		header.write(byteBuffer);

		if (isLeaf()) {
			BaseStructureHeader header = getHeader();
			switch (header.getDataTypeEnum()) {
			case DOUBLE64:
				double doubledata[] = getDoubleData();
				if (doubledata != null) {
					for (double value : doubledata) {
						byteBuffer.putDouble(value);
					}
				}
				break;

			case FLOAT32:
				float floatdata[] = getFloatData();
				if (floatdata != null) {
					for (float value : floatdata) {
						byteBuffer.putFloat(value);
					}
				}
				break;

			case LONG64:
			case ULONG64:
				long longdata[] = getLongData();
				if (longdata != null) {
					for (long value : longdata) {
						byteBuffer.putLong(value);
					}
				}
				break;

			case INT32:
			case UINT32:
				int intdata[] = getIntData();
				if (intdata != null) {
					for (int value : intdata) {
						byteBuffer.putInt(value);
					}
				}
				break;

			case SHORT16:
			case USHORT16:
				short shortdata[] = getShortData();
				if (shortdata != null) {
					for (short value : shortdata) {
						byteBuffer.putShort(value);
					}

					// might have to pad to 4 byte boundary
                    if (shortdata.length % 2 > 0) {
                        byteBuffer.putShort((short) 0);
                    }
				}
				break;

			case CHAR8:
			case UCHAR8:
            case CHARSTAR8: // rawbytes contains ascii
				byte bytedata[] = getByteData();
				if (bytedata != null) {
					for (byte value : bytedata) {
						byteBuffer.put(value);
					}

					// might have to pad to 4 byte boundary
                    byte[] bpad = {0,0,0};
                    int[]  pads = {0,3,2,1};
                    byteBuffer.put(bpad, 0, pads[bytedata.length%4]);
				}
				break;

			} // switch

		} // isLeaf
		else {
			for (BaseStructure child : children) {
				child.write(byteBuffer);
			}
		} // not leaf
		return byteBuffer.position() - startPos;
	}
	
	/**
	 * Appends int data to the structure. If the structure has no data, then this
	 * is the same as setting the data.
	 * @param data the int data to append, or set if there is no existing data.
	 * @throws EvioException
	 */
	public void appendIntData(int data[]) throws EvioException {
		
		// make sure the structure is set to hold this kind of data
		DataType dataType = header.getDataTypeEnum();
		if ((dataType != DataType.INT32) && (dataType != DataType.UINT32)) {
			throw new EvioException("Tried to append int data to a structure of type: " + dataType);
		}
		
		// if no data to append, just cave
		if (data == null) {
			return;
		}
		
		// if no existing data, life is easy
		if (intData == null) {
			intData = data;
            numberDataItems = data.length;
		}
		else {
			int size1 = intData.length; // existing data
			int size2 = data.length;    // new data to append
			intData = Arrays.copyOf(intData, size1 + size2);  // expand existing array
			System.arraycopy(data, 0, intData, size1, size2); // append
            numberDataItems += data.length;
		}
		setAllHeaderLengths();
	}
	
	/**
	 * Appends short data to the structure. If the structure has no data, then this
	 * is the same as setting the data.
	 * @param data the short data to append, or set if there is no existing data.
	 * @throws EvioException
	 */
	public void appendShortData(short data[]) throws EvioException {
		
		// make sure the structure is set to hold this kind of data
		DataType dataType = header.getDataTypeEnum();
		if ((dataType != DataType.SHORT16) && (dataType != DataType.USHORT16)) {
			throw new EvioException("Tried to append short data to a structure of type: " + dataType);
		}
		
		// if no data to append, just cave
		if (data == null) {
			return;
		}
		
		// if no existing data, life is easy
		if (shortData == null) {
			shortData = data;
            numberDataItems = data.length;
		}
		else {
			int size1 = shortData.length; // existing data
			int size2 = data.length;      // new data to append
			shortData = Arrays.copyOf(shortData, size1 + size2); // expand existing array
			System.arraycopy(data, 0, shortData, size1, size2);  // append
            numberDataItems += data.length;
		}
		setAllHeaderLengths();
	}

	
	/**
	 * Appends long data to the structure. If the structure has no data, then this
	 * is the same as setting the data.
	 * @param data the long data to append, or set if there is no existing data.
	 * @throws EvioException
	 */
	public void appendLongData(long data[]) throws EvioException {
		
		// make sure the structure is set to hold this kind of data
		DataType dataType = header.getDataTypeEnum();
		if ((dataType != DataType.LONG64) && (dataType != DataType.ULONG64)) {
			throw new EvioException("Tried to append long data to a structure of type: " + dataType);
		}
		
		// if no data to append, just cave
		if (data == null) {
			return;
		}
		
		// if no existing data, life is easy
		if (longData == null) {
			longData = data;
            numberDataItems = data.length;
		}
		else {
			int size1 = longData.length; // existing data
			int size2 = data.length;     // new data to append
			longData = Arrays.copyOf(longData, size1 + size2); // expand existing array
			System.arraycopy(data, 0, longData, size1, size2); // append
            numberDataItems += data.length;
		}
		setAllHeaderLengths();
	}
	
	/**
	 * Appends byte data to the structure. If the structure has no data, then this
	 * is the same as setting the data.
	 * @param data the byte data to append, or set if there is no existing data.
	 * @throws EvioException
	 */
	public void appendByteData(byte data[]) throws EvioException {
		
		// make sure the structure is set to hold this kind of data
		DataType dataType = header.getDataTypeEnum();
		if ((dataType != DataType.CHAR8) && (dataType != DataType.UCHAR8)) {
			throw new EvioException("Tried to append byte data to a structure of type: " + dataType);
		}
		
		// if no data to append, just cave
		if (data == null) {
			return;
		}
		
		// if no existing data, life is easy
		if (rawBytes == null) {
			rawBytes = data;
            numberDataItems = data.length;
		}
		else {
			int size1 = rawBytes.length; // existing data
			int size2 = data.length;     // new data to append
			rawBytes = Arrays.copyOf(rawBytes, size1 + size2); // expand existing array
			System.arraycopy(data, 0, rawBytes, size1, size2); // append
            numberDataItems += data.length;
		}
		setAllHeaderLengths();
	}
	
	/**
	 * Appends float data to the structure. If the structure has no data, then this
	 * is the same as setting the data.
	 * @param data the float data to append, or set if there is no existing data.
	 * @throws EvioException
	 */
	public void appendFloatData(float data[]) throws EvioException {
		
		// make sure the structure is set to hold this kind of data
		DataType dataType = header.getDataTypeEnum();
		if (dataType != DataType.FLOAT32) {
			throw new EvioException("Tried to append float data to a structure of type: " + dataType);
		}
		
		// if no data to append, just cave
		if (data == null) {
			return;
		}
		
		// if no existing data, life is easy
		if (floatData == null) {
			floatData = data;
            numberDataItems = data.length;
		}
		else {
			int size1 = floatData.length; // existing data
			int size2 = data.length;      // new data to append
			floatData = Arrays.copyOf(floatData, size1 + size2); // expand existing array
			System.arraycopy(data, 0, floatData, size1, size2);  // append
            numberDataItems += data.length;
		}
		setAllHeaderLengths();
	}
	
//    /**
//     * Appends char data to the structure (as ascii). If the structure has no data, then this
//     * is the same as setting the data. Be sure to write an ending null to complete writing a
//     * properly formatted string.
//     * @param data the char data to append (as ascii), or set if there is no existing data.
//     * @throws EvioException
//     */
//    public void appendCharData(char data[]) throws EvioException {
//
//        // make sure the structure is set to hold this kind of data
//        DataType dataType = header.getDataTypeEnum();
//        if (dataType != DataType.CHARSTAR8) {
//            throw new EvioException("Tried to append char data to a structure of type: " + dataType);
//        }
//
//        // if no data to append, just cave
//        if (data == null) {
//            return;
//        }
//
//        // if no existing data, create some storage
//        if (stringData == null) {
//            stringData = new StringBuffer(data.length);
//        }
//
//        stringData.append(data);
//        try {
//            rawBytes = stringData.toString().getBytes("US-ASCII");
//        }
//        catch (UnsupportedEncodingException e) { /* will never happen */ }
//
//        setAllHeaderLengths();
//    }


    /**
     * Appends string to the structure (as ascii). If the structure has no data, then this
     * is the same as setting the data.
     * @param s the string to append (as ascii), or set if there is no existing data.
     * @throws EvioException
     */
    public void appendStringData(String s) throws EvioException {

        // make sure the structure is set to hold this kind of data
        DataType dataType = header.getDataTypeEnum();
        if (dataType != DataType.CHARSTAR8) {
            throw new EvioException("Tried to append string to a structure of type: " + dataType);
        }

        // if no data to append, just cave
        if (s == null) {
            return;
        }

        // if no existing data, create some storage
        if (stringData == null) {
            stringData = new StringBuffer(s.length() + 1);
            numberDataItems = 1;
        }
        // else remove any existing padding
        else {
            stringData.delete(stringEnd, stringData.length());
            numberDataItems++;
        }

        // add string
        stringData.append(s);

        // add ending null
        stringData.append('\000');

        // mark end of data before adding padding
        stringEnd = stringData.length();

        // Add any necessary padding to 4 byte boundaries.
        // IMPORTANT: There must be at least one '\004'
        // character at the end. This distinguishes evio
        // string array version from earlier version.
        int[] pads = {4,3,2,1};
        switch (pads[stringData.length()%4]) {
            case 4:
                stringData.append("\004\004\004\004");
                break;
            case 3:
                stringData.append("\004\004\004");
                break;
            case 2:
                stringData.append("\004\004");
                break;
            case 1:
                stringData.append('\004');
            default:
        }

        // set raw data
        try {
            rawBytes = stringData.toString().getBytes("US-ASCII");
        }
        catch (UnsupportedEncodingException e) { /* will never happen */ }

        setAllHeaderLengths();
    }


    /**
	 * Appends double data to the structure. If the structure has no data, then this
	 * is the same as setting the data.
	 * @param data the double data to append, or set if there is no existing data.
	 * @throws EvioException
	 */
	public void appendDoubleData(double data[]) throws EvioException {
		
		// make sure the structure is set to hold this kind of data
		DataType dataType = header.getDataTypeEnum();
		if (dataType != DataType.DOUBLE64) {
			throw new EvioException("Tried to append double data to a structure of type: " + dataType);
		}
		
		// if no data to append, just cave
		if (data == null) {
			return;
		}
		
		// if no existing data, life is easy
		if (doubleData == null) {
			doubleData = data;
            numberDataItems = data.length;
		}
		else {
			int size1 = doubleData.length; // existing data
			int size2 = data.length;       // new data to append
			doubleData = Arrays.copyOf(doubleData, size1 + size2); // expand existing array
			System.arraycopy(data, 0, doubleData, size1, size2);   // append
            numberDataItems += data.length;
		}
		setAllHeaderLengths();
	}


}
