package org.jlab.coda.jevio;

import java.nio.*;

/**
 * This utility class contains methods for transforming a raw byte array into arrays of
 * other types such as int and double. It also contains a method for byte swapping evio
 * format data as well as other handy methods.
 * 
 * @author heddle
 * @author timmer
 * @author wolin
 *
 */
public class ByteDataTransformer {

	/**
	 * Converts a byte array into an int array.
	 * 
     * @param bytes the byte array.
     * @param byteOrder the endianness of the data in the byte array,
     *       {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
	 * @return the raw bytes converted into an int array.
	 */
	public static int[] getAsIntArray(byte bytes[], ByteOrder byteOrder) {
		ByteBuffer byteBuffer = ByteBuffer.wrap(bytes).order(byteOrder);
		int intsize = bytes.length / 4;
		int array[] = new int[intsize];
		for (int i = 0; i < intsize; i++) {
			array[i] = byteBuffer.getInt();
		}
		return array;
	}
	
	/**
	 * Converts a byte array into an short array.
	 * 
	 * @param bytes the byte array.
     * @param byteOrder the endianness of the data in the byte array,
     *       {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
	 * @return the raw bytes converted into an int array.
	 */
	public static short[] getAsShortArray(byte bytes[], ByteOrder byteOrder) {
		ByteBuffer byteBuffer = ByteBuffer.wrap(bytes).order(byteOrder);
		int shortsize = bytes.length / 2;
		short array[] = new short[shortsize];
		for (int i = 0; i < shortsize; i++) {
			array[i] = byteBuffer.getShort();
		}
		return array;
	}

	
	/**
	 * Converts a byte array into a long array.
	 * 
	 * @param bytes the byte array.
     * @param byteOrder the endianness of the data in the byte array,
     *       {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
	 * @return the raw bytes converted into a long array.
	 */
	public static long[] getAsLongArray(byte bytes[], ByteOrder byteOrder) {
		ByteBuffer byteBuffer = ByteBuffer.wrap(bytes).order(byteOrder);
		int longsize = bytes.length / 8;
		long array[] = new long[longsize];
		for (int i = 0; i < longsize; i++) {
			array[i] = byteBuffer.getLong();
		}
		return array;
	}

	/**
	 * Converts a byte array into a double array.
	 * 
	 * @param bytes the byte array.
     * @param byteOrder the endianness of the data in the byte array,
     *       {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
	 * @return the raw bytes converted into a double array.
	 */
	public static double[] getAsDoubleArray(byte bytes[], ByteOrder byteOrder) {
		ByteBuffer byteBuffer = ByteBuffer.wrap(bytes).order(byteOrder);
		int doublesize = bytes.length / 8;
		double array[] = new double[doublesize];
		for (int i = 0; i < doublesize; i++) {
			array[i] = byteBuffer.getDouble();
		}
		return array;
	}
	
	/**
	 * Converts a byte array into a float array.
	 * 
	 * @param bytes the byte array.
     * @param byteOrder the endianness of the data in the byte array,
     *       {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
	 * @return the raw bytes converted into a float array.
	 */
	public static float[] getAsFloatArray(byte bytes[], ByteOrder byteOrder) {
		ByteBuffer byteBuffer = ByteBuffer.wrap(bytes).order(byteOrder);
		int floatsize = bytes.length / 4;
		float array[] = new float[floatsize];
		for (int i = 0; i < floatsize; i++) {
			array[i] = byteBuffer.getFloat();
		}
		return array;
	}

    /**
     * Copies an integer value into 4 bytes of a byte array.
     * @param intVal integer value
     * @param b byte array
     * @param off offset into the byte array
     */
    public static final void intToBytes(int intVal, byte[] b, int off) {
        b[off]   = (byte) ((intVal & 0xff000000) >>> 24);
        b[off+1] = (byte) ((intVal & 0x00ff0000) >>> 16);
        b[off+2] = (byte) ((intVal & 0x0000ff00) >>>  8);
        b[off+3] = (byte)  (intVal & 0x000000ff);
    }


    /**
     * Copies a short value into 2 bytes of a byte array.
     * @param val short value
     * @param b byte array
     * @param off offset into the byte array
     */
    public static final void shortToBytes(short val, byte[] b, int off) {
      b[off]   = (byte) ((val & 0x0000ff00) >>>  8);
      b[off+1] = (byte)  (val & 0x000000ff);
    }


    /**
     * Copies a 2-byte short value into a 4-byte int while preserving bit pattern.
     * In other words, it treats the short as an unsigned value. The resulting int
     * does NOT undergo sign extension (bits of highest 2 bytes are not all ones),
     * and will not be a negative number.
     *
     * @param shortVal short value considered to be unsigned
     * @return the equivalent int value
     */
    public static final int shortBitsToInt(short shortVal) {
        return (shortVal & 0x0000ffff);
    }


    /**
     * Copies a 1-byte byte value into a 4-byte int while preserving bit pattern.
     * In other words, it treats the byte as an unsigned value. The resulting int
     * does NOT undergo sign extension (bits of highest 3 bytes are not all ones),
     * and will not be a negative number.
     *
     * @param byteVal byte value considered to be unsigned
     * @return the equivalent int value
     */
    public static final int byteBitsToInt(byte byteVal) {
        return (byteVal & 0x000000ff);
    }


    //--------------------------------
    // EVIO FORMAT DATA SWAPPING
    //--------------------------------

    /**
     * Swaps evio format data (event or bank) contained in a ByteBuffer.
     * NOTE: the byte order of the source buffer must be set properly by
     * the caller using {@link ByteBuffer#order(java.nio.ByteOrder)}.
     *
     * @param srcBuf   evio data (event or bank) to be swapped
     * @param destBuf  destination for swapped evio data; if null, swap in place
     * @throws EvioException if data not in evio format
     */
    public static void swap(ByteBuffer srcBuf, ByteBuffer destBuf) throws EvioException {
        ByteOrder srcOrder  = srcBuf.order();
        ByteOrder destOrder = ByteOrder.BIG_ENDIAN;

        if (srcOrder == ByteOrder.BIG_ENDIAN) {
            destOrder = ByteOrder.LITTLE_ENDIAN;
        }

        // writing into this buffer will automatically swap bytes
        if (destBuf != null) {
            destBuf.order(destOrder);
        }

        try {
            swapBank(srcBuf, destBuf);
        }
        catch (BufferOverflowException e) {
            throw new EvioException("Data not in evio format", e);
        }
        catch (BufferUnderflowException e) {
            throw new EvioException("Data not in evio format", e);
        }
        catch (IndexOutOfBoundsException e) {
            throw new EvioException("Data not in evio format", e);
        }

        // if swapping in place, set buffer's new byte order now that it's completely swapped
        if (destBuf == null) {
            srcBuf.order(destOrder);
        }
    }

    /**
     * Swaps evio event or bank data.
     *
     * @param bank  evio data (event or bank) to be swapped
     * @return bank object with swapped data
     * @throws EvioException  if data not in evio format
     */
    public static EvioEvent swap(EvioBank bank) throws EvioException {
        ByteOrder newOrder = (bank.getByteOrder() == ByteOrder.BIG_ENDIAN) ? ByteOrder.LITTLE_ENDIAN : ByteOrder.BIG_ENDIAN;

        byte[] array = bank.toArray();
        ByteBuffer buf = ByteBuffer.wrap(array);
        buf.order(bank.getByteOrder());

        try {
            swapBank(buf, null);
        }
        catch (BufferOverflowException e) {
            throw new EvioException("Data not in evio format", e);
        }
        catch (BufferUnderflowException e) {
            throw new EvioException("Data not in evio format", e);
        }
        catch (IndexOutOfBoundsException e) {
            throw new EvioException("Data not in evio format", e);
        }

        ByteParser parser = new ByteParser();
        buf.rewind();
        // set buf to new byte order now that it's completely swapped
        buf.order(newOrder);
        return parser.parseEvent(buf);
    }

    /**
     * Copies buffer of 4-byte integers.
     *
     * @param srcBuf   input data buffer
     * @param length   number of integer words to copy
     * @param destBuf  destination data buffer; if null, swap in place
     */
    private static void copyData(ByteBuffer srcBuf, int length, ByteBuffer destBuf) {

        if (destBuf == null) return;

        int numBytes = 4*length;
        byte[] bytes = new byte[numBytes];
        
        srcBuf.get(bytes, 0, numBytes);
        destBuf.put(bytes, 0, numBytes);

        return;
    }

    /**
     * Swaps buffer of 4-byte integers.
     *
     * @param srcBuf   input data buffer
     * @param length   number of integer words to swap
     * @param destBuf  destination data buffer; if null, swap in place
     */
    private static void swapIntArray(ByteBuffer srcBuf, int length, ByteBuffer destBuf) {

        if (destBuf == null) {
            int pos;
            for (int i = 0; i < length; i++) {
                pos = srcBuf.position();
                srcBuf.putInt(Integer.reverseBytes(srcBuf.getInt(pos)));
            }
        }
        else {
            for (int i = 0; i < length; i++) {
                destBuf.putInt(srcBuf.getInt());
            }
        }
        return;
    }

    /**
     * Swaps 8-byte longs packed into 4-byte integer buffer.
     *
     * @param srcBuf   input data buffer
     * @param length   number of integer words to swap
     * @param destBuf  destination data buffer; if null, swap in place
     */
    private static void swapLongArray(ByteBuffer srcBuf, int length, ByteBuffer destBuf) {
        int numLongs = length/2;

        if (destBuf == null) {
            int pos;
            for (int i = 0; i < numLongs; i++) {
                pos = srcBuf.position();
                srcBuf.putLong(Long.reverseBytes(srcBuf.getLong(pos)));
            }
        }
        else {
            for (int i = 0; i < numLongs; i++) {
                destBuf.putLong(srcBuf.getLong());
            }
        }
        return;
    }

    /**
     * Swaps 2-byte shorts packed in 4-byte integer buffer.
     *
     * @param srcBuf   input data buffer
     * @param length   number of integer words to swap
     * @param destBuf  destination data buffer; if null, swap in place
     */
    private static void swapShortArray(ByteBuffer srcBuf, int length, ByteBuffer destBuf) {
        int numShorts = 2*length;

        if (destBuf == null) {
            int pos;
            for (int i = 0; i < numShorts; i++) {
                pos = srcBuf.position();
                srcBuf.putShort(Short.reverseBytes(srcBuf.getShort(pos)));
            }
        }
        else {
            for (int i = 0; i < numShorts; i++) {
                destBuf.putShort(srcBuf.getShort());
            }
        }
        return;
    }

    /**
     * Swap an evio bank.
     *
     * @param srcBuf   evio bank data to swap
     * @param destBuf  destination buffer for swapped bank; if null, swap in place
     * @return size of data (in 4 byte ints) contained in bank plus header
     * @throws EvioException if data not in evio format
     */
    private static int swapBank(ByteBuffer srcBuf, ByteBuffer destBuf) throws EvioException {

        // parse 2 header ints to find data length & type
        srcBuf.mark();
        int dataLength = srcBuf.getInt() - 1;
        DataType dataType = DataType.getDataType((srcBuf.getInt() >> 8) & 0xff);
        srcBuf.reset();

        // swap 2 header ints
        swapIntArray(srcBuf, 2, destBuf); // does not advance position

        // swap data following header
        swapData(srcBuf, dataType, dataLength, destBuf);

        // return total length of this bank including both header ints
        return dataLength + 2;
    }

    /**
     * Swap an evio segment.
     *
     * @param srcBuf   byte buffer to swap
     * @param destBuf  destination buffer for swapped event; if null, swap in place
     * @return size of data (in 4 byte ints) contained in segment plus header
     * @throws EvioException if data not in evio format
     */
    private static int swapSegment(ByteBuffer srcBuf, ByteBuffer destBuf) throws EvioException {

        // parse 1 header int to find data length & type
        srcBuf.mark();
        int d = srcBuf.getInt();
        srcBuf.reset();
        int dataLength = (d & 0xffff);
        DataType dataType = DataType.getDataType((d >> 16) & 0xff);

        // swap 1 header int
        swapIntArray(srcBuf, 1, destBuf);

        // swap data following header
        swapData(srcBuf, dataType, dataLength, destBuf);

        // return total length of this bank including header int
        return dataLength + 1;
    }


    /**
     * Swap an evio tagsegment.
     *
     * @param srcBuf   byte buffer to swap
     * @param destBuf  destination buffer for swapped event; if null, swap in place
     * @return size of data (in 4 byte ints) contained in tagsegment plus header
     * @throws EvioException if data not in evio format
     */
    private static int swapTagSegment(ByteBuffer srcBuf, ByteBuffer destBuf) throws EvioException {

        // parse 1 header int to find data length & type
        srcBuf.mark();
        int d = srcBuf.getInt();
        srcBuf.reset();
        int dataLength = (d & 0xffff);
        DataType dataType = DataType.getDataType((d >> 16) & 0xf);

        // swap 1 header int
        swapIntArray(srcBuf, 1, destBuf);

        // swap data following header
        swapData(srcBuf, dataType, dataLength, destBuf);

        // return total length of this bank including header int
        return dataLength + 1;
    }


    /**
     * Handle the swapping of evio data or container of evio data.
     *
     * @param srcBuf   byte buffer to swap
     * @param type     type of data in the data buffer
     * @param length   length of data (4 byte ints) to swap
     * @param destBuf  destination buffer for swapped event; if null, swap in place
     * @throws EvioException if data not in evio format
     */
    private static void swapData(ByteBuffer srcBuf, DataType type, int length,
                                 ByteBuffer destBuf) throws EvioException {

        int len = 0;

        // A container (bank, segment, or tagsegment) to be swapped has called this method.
        // We switch on the type of data or containers held by this top-level container.
        // If data is held, just swap it here or call swap<Fragment> methods if containers
        // are held.

        switch (type) {

            // 32 bit swap
            case INT32:
            case UINT32:
            case FLOAT32:
                swapIntArray(srcBuf, length, destBuf);
                break;

            // 16 bit swap
            case SHORT16:
            case USHORT16:
                swapShortArray(srcBuf, length, destBuf);
                break;

            // 64 bit swap
            case LONG64:
            case ULONG64:
            case DOUBLE64:
                swapLongArray(srcBuf, length, destBuf);
                break;

            // no swap
            case UNKNOWN32:
            case CHAR8:
            case UCHAR8:
            case CHARSTAR8:
                copyData(srcBuf, length, destBuf);
                break;

            // container of banks
            case BANK:
            case ALSOBANK:
                // The calling container is a bank of banks and may contain
                // multiple banks, each of which need to be swapped.
                while (len < length) {
                    len += swapBank(srcBuf, destBuf);
                }
                if (len != length) {
                    throw new EvioException("Data not in evio format");
                }
                break;

            // container of segments
            case SEGMENT:
            case ALSOSEGMENT:
                while (len < length) {
                    len += swapSegment(srcBuf, destBuf);
                }
                if (len != length) {
                    throw new EvioException("Data not in evio format");
                }
                break;

            // container of tagsegments
            case TAGSEGMENT:
            case ALSOTAGSEGMENT:
                while (len < length) {
                    len += swapTagSegment(srcBuf, destBuf);
                }
                if (len != length) {
                    throw new EvioException("Data not in evio format");
                }
                break;

            // unknown type, just copy
            default:
                copyData(srcBuf, length, destBuf);
                break;
        }

        return;
    }

}
