package org.jlab.coda.jevio;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.io.DataInputStream;
import java.io.IOException;

import javax.swing.event.EventListenerList;
import javax.swing.tree.DefaultTreeModel;

/**
 * This class controls the creating of an event by the parsing of a byte array.
 * Unlike the EventParser class, it does not use a singleton pattern because the parsing of bytes is used primarily
 * for reading data sent over the network or from an ET system and is used in a multithreaded environment.
 * This class acts like EvioFile and EventParser together, except the data in our case is obtained by reading
 * IO streams (eg. from a socket). Thus, it's much simpler than reading a file and the result is we have a
 * byte array to parse.<p>
 *
 * @author timmer
 */
public class ByteParser {

    /**
     * Listener list for structures (banks, segments, tagsegments) encountered while processing an event.
     */
    private EventListenerList evioListenerList;

    /**
     * This flag determines whether notification of listeners is active. Normally it is. But in some cases it should be
     * temporarily suspended. For example, in a "goto event" process, the listeners will not be notified of the
     * intervening events as the file is scanned to get to the target event.
     */
    protected boolean notificationActive = true;

    /**
      * A single global structure filter. This may prove to be a limitation: perhaps we want one per listener.
      */
    private IEvioFilter evioFilter;

    /**
     * No-arg constructor.
     */
    public ByteParser() { }


    /**
     * This is the workhorse method for parsing the byte array. It will drill down and uncover all structures
     * (banks, segments, and tagsegments) and notify any interested listeners in a SAX-Like manner. <br>
     * Note: applications may choose not to provide a listener. In that case, when the event is parsed, its structures
     * may be accessed through the event's tree model, i.e., via <code>event.getTreeModel()</code>.
     *
     * @param bytes     the binary data to be parsed.
     * @param byteOrder the byte order of the binary array, either
     *        {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
     * @return an EvioEvent object created from the byte array.
     * @throws EvioException if data not in evio format.
     */
    public EvioEvent parseEvent(byte[] bytes, ByteOrder byteOrder) throws EvioException {
        // wrap the byte array in a ByteBuffer for ease of parsing & automatic byte swapping
        ByteBuffer buf = ByteBuffer.wrap(bytes).order(byteOrder);
        return parseEvent(buf);
    }


    /**
     * This is the workhorse method for parsing the byte buffer. It will drill down and uncover all structures
     * (banks, segments, and tagsegments) and notify any interested listeners in a SAX-Like manner. <br>
     * Note: applications may choose not to provide a listener. In that case, when the event is parsed, its structures
     * may be accessed through the event's tree model, i.e., via <code>event.getTreeModel()</code>.
     *
     * @param buf buffer of binary data to parse.
     * @return an EvioEvent object created from the byte array.
     * @throws EvioException if data not in evio format.
     */
    public EvioEvent parseEvent(ByteBuffer buf) throws EvioException {
        return (EvioEvent) parseBank(buf, null, null);
    }


    /**
     * This is a method for reading data over a DataInputStream and parsing it into an EvioEvent.
     *
     * @param in data input stream containing binary data to parse in network byte order.
     * @param byteOrder the byte order of the input bytes, either
     *        {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
     * @return an EvioEvent object created from the data input stream.
     * @throws IOException if IO error.
     * @throws EvioException if data not in evio format.
     */
    public EvioEvent readEvent(DataInputStream in, ByteOrder byteOrder) throws IOException, EvioException {
        int length, lengthRead;
        length = lengthRead = in.readInt();
        if (byteOrder == ByteOrder.LITTLE_ENDIAN) {
            length = Integer.reverseBytes(lengthRead);
        }

        byte[] buffer = new byte[4*(length + 1)];

        in.readFully(buffer, 4, 4*length);

        // put length at beginning of array
        ByteDataTransformer.intToBytes(lengthRead, buffer, 0);

        return parseEvent(buffer, byteOrder);
    }


    /**
     * Parse a ByteBuffer as an evio bank.
     *
     * @param buf buffer of binary data to parse.
     * @param event the EvioEvent object being created.
     * @param parent the parent of the bank being created.
     * @return the EvioBank object created from the byte buffer.
     * @throws EvioException if data not in evio format.
     */
    private BaseStructure parseBank(ByteBuffer buf, EvioEvent event, BaseStructure parent)
            throws EvioException {

        // create & fill bank header
        BankHeader header = createBankHeader(buf);

        // create bank
        EvioBank bank;

        // if no event created yet, we're top level so create it
        if (event == null) {
            // create event which is top-level bank
            event = new EvioEvent(header);
            event.treeModel = new DefaultTreeModel(event);
            bank = event;
        }
        else {
            // create bank and put it in parent
            bank = new EvioBank(header);
            bank.setParent(parent);
            event.insert(bank, parent);
        }

        return parseStructure(buf, event, bank);
    }



    /**
     * Parse a ByteBuffer as an evio segment.
     *
     * @param buf buffer of binary data to parse.
     * @param event the EvioEvent object being created.
     * @param parent the parent of the segment being created.
     * @return the EvioSegment object created from the byte buffer.
     * @throws EvioException if data not in evio format.
     */
    private BaseStructure parseSegment(ByteBuffer buf, EvioEvent event, BaseStructure parent)
            throws EvioException {

        // create & fill segment header
        SegmentHeader header = createSegmentHeader(buf);

        // create segment and put it in parent
        EvioSegment segment = new EvioSegment(header);
        segment.setParent(parent);
        event.insert(segment, parent);

        return parseStructure(buf, event, segment);
    }


    /**
     * Parse a ByteBuffer as an evio tag segment.
     *
     * @param buf buffer of binary data to parse.
     * @param event the EvioEvent object being created.
     * @param parent the parent of the tag segment being created.
     * @return the EvioTagSegment object created from the byte buffer.
     * @throws EvioException if data not in evio format.
     */
    private BaseStructure parseTagSegment(ByteBuffer buf, EvioEvent event, BaseStructure parent)
            throws EvioException {

        // create & fill tag segment header
        TagSegmentHeader header = createTagSegmentHeader(buf);

        // create tag segment and put it in parent
        EvioTagSegment tagSegment = new EvioTagSegment(header);
        tagSegment.setParent(parent);
        event.insert(tagSegment, parent);

        return parseStructure(buf, event, tagSegment);
    }


    /**
     * Parse the evio structure (bank, segment, or tagsegment).
     *
     * @param buf buffer of binary data to parse.
     * @param event the EvioEvent object being created.
     * @param structure the structure object being created (bank, segment, or tagsegment)
     * @return object passed in as structure
     * @throws EvioException if data not in evio format.
     */
    private BaseStructure parseStructure(ByteBuffer buf, EvioEvent event, BaseStructure structure)
            throws EvioException {

        BaseStructureHeader header = structure.getHeader();

        // save position (before raw data) for returning to later
        buf.mark();

        // length of non-header (raw) data in 32-bit ints
        int dataLen = header.getLength() + 1 - header.getHeaderLength();

        // store raw bytes in structure
        byte[] bytes = new byte[4*dataLen];
        buf.get(bytes, 0, bytes.length);
        structure.setRawBytes(bytes);
        structure.setByteOrder(buf.order());

        // Does the present structure contain structures? (as opposed to a leaf, which contains primitives). If
        // it is a leaf (condition == true), we are done. That will leave the raw bytes in the "leaf" structures
        // (e.g., a bank of ints) -- which will be interpreted by the various "get data" methods.
        if (!header.getDataTypeEnum().isStructure()) {
            // notify the listeners
            notifyEvioListeners(event, structure);
            return structure;
        }

        // Move buffer position before raw data since the raw
        // data is an evio container and needs to be parsed.
        buf.reset();

        // parse structures contained in raw data
        parseData(buf, header.getDataTypeEnum(), dataLen, event, structure);

        // return same structure passed in
        return structure;
    }


    /**
     * Create a bank header from the next eight bytes of the byte buffer.
     *
     * @param byteBuffer the byte buffer, probably from a bank that encloses this new bank.
     * @return the new header.
     */
    private BankHeader createBankHeader(ByteBuffer byteBuffer) {
        BankHeader header = new BankHeader();
        header.setLength(byteBuffer.getInt());

        if (byteBuffer.order() == ByteOrder.BIG_ENDIAN) {
            // interested in bit pattern, not negative numbers
            header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
            header.setDataType(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
        }
        else {
            header.setNumber(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            header.setDataType(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            header.setTag(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
        }
        return header;
    }


    /**
     * Create a segment header from the next four bytes of the byte buffer.
     *
     * @param byteBuffer the byte buffer, probably from a bank that encloses this new segment.
     * @return the new header.
     */
    private SegmentHeader createSegmentHeader(ByteBuffer byteBuffer) {
        SegmentHeader header = new SegmentHeader();

        if (byteBuffer.order() == ByteOrder.BIG_ENDIAN) {
            header.setTag(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            header.setDataType(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            header.setLength(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
        }
        else {
            header.setLength(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
            header.setDataType(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
            header.setTag(ByteDataTransformer.byteBitsToInt(byteBuffer.get()));
        }

        return header;
    }


    /**
     * Create a tag segment header from the next four bytes of the byte buffer.
     *
     * @param byteBuffer the byte buffer, probably from a bank that encloses this new tag segment.
     * @return the new header.
     */
    private TagSegmentHeader createTagSegmentHeader(ByteBuffer byteBuffer) {
        TagSegmentHeader header = new TagSegmentHeader();

        int temp;

        if (byteBuffer.order() == ByteOrder.BIG_ENDIAN) {
            temp = ByteDataTransformer.shortBitsToInt(byteBuffer.getShort());
            header.setTag(temp >>> 4);
            header.setDataType(temp & 0xF);
            header.setLength(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
        }
        else {
            header.setLength(ByteDataTransformer.shortBitsToInt(byteBuffer.getShort()));
            temp = ByteDataTransformer.shortBitsToInt(byteBuffer.getShort());
            header.setTag(temp >>> 4);
            header.setDataType(temp & 0xF);
        }

        return header;
    }


    /**
     * Handle parsing raw data bytes into containers of data.
     *
     * @param buf    the buffer of binary data to parse.
     * @param type   the type of container in the data buffer.
     * @param length the length of data in the data buffer to be parsed for these containers.
     * @param event  the EvioEvent object being created.
     * @param parent the parent of the structure (banks, segments, or tagsegments) being created.
     * @throws EvioException if data not in evio format.
     */
    private void parseData(ByteBuffer buf, DataType type, int length, EvioEvent event, BaseStructure parent)
            throws EvioException {

        int len = 0;

        switch (type) {

            // parent is container of banks
            case BANK:
            case ALSOBANK:
                // The calling container is a bank of banks and may contain
                // multiple banks, each of which need to be parsed.
                while (len < length) {
                    BaseStructure bank = parseBank(buf, event, parent);
                    len += bank.getHeader().getLength() + 1;
                }
                if (len != length) {
                    throw new EvioException("Data not in evio format");
                }

                break;

            // parent is container of segments
            case SEGMENT:
            case ALSOSEGMENT:
                while (len < length) {
                    BaseStructure seg = parseSegment(buf, event, parent);
                    len += seg.getHeader().getLength() + 1;
                }
                if (len != length) {
                    throw new EvioException("Data not in evio format");
                }
                break;

            // parent is container of tagsegments
            case TAGSEGMENT:
            case ALSOTAGSEGMENT:
                while (len < length) {
                    BaseStructure tag = parseTagSegment(buf, event, parent);
                    len += tag.getHeader().getLength() + 1;
                }
                if (len != length) {
                    throw new EvioException("Data not in evio format");
                }
                break;

            // non-container type, do nothing
            default:
                break;
        }

        // notify the listeners
        notifyEvioListeners(event, parent);

        return;
    }


    /**
     * This is when a structure is encountered while parsing an event. It notifies all listeners about the structure.
     *
     * @param structure the structure encountered, which may be a Bank, Segment, or TagSegment.
     */
    protected void notifyEvioListeners(EvioEvent evioEvent, IEvioStructure structure) {

        // are notifications turned off?
        if (!notificationActive) {
            return;
        }

        if (evioListenerList == null) {
            return;
        }

        // if there is a global filter, lets use it.
        if (evioFilter != null) {
            if (!evioFilter.accept(structure.getStructureType(), structure.getHeader())) {
                return;
            }
        }

        // Guaranteed to return a non-null array
        Object[] listeners = evioListenerList.getListenerList();

        // This weird loop is the bullet proof way of notifying all listeners.
        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            if (listeners[i] == IEvioListener.class) {
                ((IEvioListener) listeners[i + 1]).gotStructure(evioEvent, structure);
            }
        }
    }

    /**
     * Remove an Evio listener. Evio listeners listen for structures encountered when an event is being parsed.
     *
     * @param listener The Evio listener to remove.
     */
    public void removeEvioListener(IEvioListener listener) {

        if ((listener == null) || (evioListenerList == null)) {
            return;
        }

        evioListenerList.remove(IEvioListener.class, listener);
    }

    /**
     * Add an Evio listener. Evio listeners listen for structures encountered when an event is being parsed.
     *
     * @param listener The Evio listener to add.
     */
    public void addEvioListener(IEvioListener listener) {

        if (listener == null) {
            return;
        }

        if (evioListenerList == null) {
            evioListenerList = new EventListenerList();
        }

        evioListenerList.add(IEvioListener.class, listener);
    }

    /**
     * Get the flag determining whether notification of listeners is active. Normally it is. But in some cases it should
     * be temporarily suspended. For example, in a "goto event" process, the listeners will not be notified of the
     * intervening events as the file is scanned to get to the target event.
     *
     * @return <code>true</code> if notification of events to the listeners is active.
     */
    public boolean isNotificationActive() {
        return notificationActive;
    }

    /**
     * Set the flag determining whether notification of listeners is active. Normally it is. But in some cases it
     * should be temporarily suspended. For example, in a "goto event" process, the listeners will not be notified of
     * the intervening events as the file is scanned to get to the target event.
     *
     * @param notificationActive set <code>true</code> if notification of events to the listeners is active.
     */
    public void setNotificationActive(boolean notificationActive) {
        this.notificationActive = notificationActive;
    }

    /**
     * Set the global filter used for filtering structures. If set to <code>null</code>, the default, then all
     * structures will be sent to the listeners.
     *
     * @param evioFilter the filter to set.
     * @see IEvioFilter
     */
    public void setEvioFilter(IEvioFilter evioFilter) {
        this.evioFilter = evioFilter;
    }


}
