package org.jlab.coda.jevio;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * An EventWriter object is used for writing events to a file in evio versions 1, 2 and 3.
 * @author heddle
 *
 */
public class EventWriter {
	
    /**
     * Offset to where the block length is written in the byte buffer, which always has a physical
     * record header at the top.
     */
    private static int BLOCK_LENGTH_OFFSET = 0;

    /**
     * Offset to where the header length is written in the byte buffer, which always has a physical
     * record header at the top.
     */
    private static int HEADER_LENGTH_OFFSET = 8;

	/**
	 * Offset to where the start is written in the byte buffer, which always has a physical
	 * record header at the top. The start is the offset to the start of the first event that
	 * starts in a block.
	 */
	private static int START_OFFSET = 12;

    /**
     * Offset to where the end is written in the byte buffer, which always has a physical
     * record header at the top. The end is the number of valid words (header + data) and will
     * be the same as the blocksize for all blocks but the last.
     */
    private static int END_OFFSET = 16;

    /**
     * Offset to where the version number is written in the byte buffer, which always has a physical
     * record header at the top.
     */
    private static int VERSION_OFFSET = 20;

    /**
     * Offset to where reserved 1 is written in the byte buffer, which always has a physical
     * record header at the top.
     */
    private static int RESERVED1_OFFSET = 24;

	/**
	 * The default size used for writing a file, in ints. Max allowed: 32768. Must be a multiple of 256.
	 */
	private static int DEFAULT_BLOCK_SIZE = 8192;
	
	/**
	 * The output stream used for writing a file
	 */
	private DataOutputStream dataOutputStream;
	
	/**
	 * Block size for anachronistic physical record (block headers). A typical number is 8192.
     * Must be a multiple of 256 with a max value of 32768.
	 */
	private int blockSize;
	
	/**
	 * This is an internal byte buffer corresponding to one block (physical record). When this gets full it
     * will be flushed to the external output stream (<code>dataOutputStream</code>) that represents the file.
	 */
	private ByteBuffer blockBuffer;
	
	/**
	 * Checks for first event started in current buffer.
	 */
	private boolean updatedStart;
	
    /**
     * Running count of the block number.
     */
    private int blockNumber = 0;
    
    /**
     * We are going to use reserved1 in the block header to hold the ordinal count
     * of the last event completed in the current block.<br>
     */
    private int lastEventWritten = 0;


    /**
     * Creates an event writer for writing to a file.
     * @param file the file to write to. Will be created.<br>
     * WARNING: If the file already exists, an attempt will be made to delete it. Appending
     * to an existing event file is not supported.
     * @throws EvioException
      */
    public EventWriter(File file) throws EvioException {
        this(file, DEFAULT_BLOCK_SIZE);
    }

    /**
     * Creates an event writer for writing to a file.
     * @param filename name of the file to write to. Will be created.<br>
     * WARNING: If the file already exists, an attempt will be made to delete it. Appending
     * to an existing event file is not supported.
     * @throws EvioException
      */
    public EventWriter(String filename) throws EvioException {
        this(new File(filename), DEFAULT_BLOCK_SIZE);
    }

    /**
     * Create an <code>EventWriter</code> for writing events to a file.
     * @param file the file to write to. Will be created.<br> 
     * WARNING: If the file already exists, an attempt will be made to delete it. Appending
     * to an existing event file is not supported.
     * @param blockSize the blocksize to use. A typical number is 8192. Must be a multiple of 256 with a
     * max value of 32768.
     * @throws EvioException
     */
    public EventWriter(File file, int blockSize) throws EvioException {
    	this.blockSize = blockSize;
    	
		try {
			dataOutputStream = new DataOutputStream(new FileOutputStream(file));
		}
		catch (FileNotFoundException e) {
			throw new EvioException(e.getMessage());
		}
    }
    
    /**
     * Create an <code>EventWriter</code> for writing events to a file.
     * @param file the file to write to. Will be created.<br> 
     * WARNING: If the file already exists, an attempt will be made to delete it. Appending
     * to an existing event file is not supported.
     * @param blockSize the blocksize to use. A typical number is 8192. Must be a multiple of 256 with a
     * max value of 32768.
     * @param okToDelete if <code>true</code> and the file already exists, an attempt will be made to delete the existing file.
     * Appending to an existing event file is not supported at this time.
     */
    public EventWriter(File file, int blockSize, boolean okToDelete) throws EvioException {
    	this.blockSize = blockSize;
    	
    	if (file == null) {
    		throw new EvioException("Null file in EventWriter constructor");
    	}
    	
    	// handle existing file
    	if (file.exists() && file.isFile()) {
    		if (okToDelete) {
    			boolean deleted = file.delete();
    			if (!deleted) {
            		throw new EvioException("File in EventWriter constructor already exists, and could not be deleted. \n" +
            				"File: " + file.getPath());
    			}
    		}
    		else { // user requested no delete
        		throw new EvioException("File in EventWriter constructor already exists, and user requested no deletion. \n" +
        				"File: " + file.getPath());
    		}
    	}
    	
    	
		try {
			dataOutputStream = new DataOutputStream(new FileOutputStream(file));
		}
		catch (FileNotFoundException e) {
    		throw new EvioException("Fow whatever reason, the file could not be opened. \n " + 
    				"It could mean there is no permission to open the file for writing \n" +
    				"or that the file was a directory." +
    				"File: " + file.getPath());
		}
    }
    
    
    /**
     * Get a clean buffer. This is called at the start and also after every physical record
     * is flushed to the file.
     */
    private void getCleanBuffer() {
    	if (blockBuffer == null) {
    		blockBuffer = ByteBuffer.allocate(4*blockSize); // apparently this buffer is backed by array
    		blockBuffer.order(ByteOrder.nativeOrder());
    	}
    	else {
    		blockBuffer.clear();
    	}

    	updatedStart = false;

    	//write header words, some of which will be overwritten later when the values are determined,
		blockBuffer.putInt(blockSize);
		blockBuffer.putInt(blockNumber++);
		blockBuffer.putInt(8);         // header size always 8
		blockBuffer.putInt(0);         // start will be updated
		blockBuffer.putInt(blockSize); // end is usually blocksize
		blockBuffer.putInt(2);         // version is 1, 2 or 3 (same format so just pick 2)
		blockBuffer.putInt(0);         // reserved1 will be updated
		blockBuffer.putInt(BlockHeader.MAGIC_NUMBER); // should be MAGIC_NUMBER
    }


    /**
     * Close the underlying data output stream, and with it the file.
     */
    public void close() {
        try {
            // write final bytes, if there are any to be written
            int position = blockBuffer.position();

            if (position > 0) {
                dataOutputStream.write(blockBuffer.array());
            }

            // finish writing file
            dataOutputStream.close();
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }


    /**
     * Write a bank to the file in evio version 1, 2, 3 format.
     *
     * @param bank the bank to write.
     * @throws EvioException
     */
    public void writeEvent(EvioBank bank) throws EvioException {
    	if (bank == null) {
    		throw new EvioException("Attempt to write null event using EventWriter");
    	}

        // if first event is being written, create buffer
        if (blockBuffer == null) {
            getCleanBuffer();
        }
    	
    	//TODO call event.setComputedLength to make sure all headers have correct length?
    	int bytelen = bank.getTotalBytes();
    	ByteBuffer eventBuffer = ByteBuffer.allocate(bytelen);
    	eventBuffer.order(ByteOrder.nativeOrder());

    	bank.write(eventBuffer);

    	// is the first event started in this buffer? If not, this is the first event that starts in this
    	// block and we have to modify the start field accordingly
    	if (!updatedStart) {
    		updateStart(blockBuffer.position() / 4);
    		updatedStart = true;
    	}
    	
    	// update the reserved1 field because another event is starting in this physical record (block)
		updateReserved1(++lastEventWritten);

		// reset to top of event buffer
		eventBuffer.position(0);

		// the big question is: is there or is there not enough space in the block buffer to finish
		// writing the event? If not we write as much as possible, flush, and repeat until we do
		// have enough space to finish writing the event.
		while (eventBuffer.remaining() > blockBuffer.remaining()) {
			updateEnd(blockSize);
			// finish filling current block buffer
			int remaining = blockBuffer.remaining();
			for (int i = 0; i < remaining; i++) {
				blockBuffer.put(eventBuffer.get());
			}
			
			//byte buffer is now full. Flush and reset
			try {
				dataOutputStream.write(blockBuffer.array());
				getCleanBuffer();
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		// now we DO have enough space for remaining, if there is any more to write.
		if (eventBuffer.remaining() > 0) {
			// we are completing an event
			blockBuffer.put(eventBuffer.array(), eventBuffer.position(), eventBuffer.remaining());
			updateEnd(blockBuffer.position() / 4);
		}
	}

    /**
     * Update the block length field in the header using an absolute put.
     * @param length the new value for block length.
     */
    private void updateBlockLength(int length) {
        blockBuffer.putInt(BLOCK_LENGTH_OFFSET, length);
    }

    /**
     * Update the start field in the header using an absolute put.
     * @param start the new value for start.
     */
    private void updateStart(int start) {
        blockBuffer.putInt(START_OFFSET, start);
    }

	/**
	 * Update the end field in the header using an absolute put.
	 * @param end the new value for end.
	 */
	private void updateEnd(int end) {
		blockBuffer.putInt(END_OFFSET, end);
	}
	
	/**
	 * Update the reserved1 field in the header using an absolute put.
	 * @param reserved1 the new value for reserved1.
	 */
	private void updateReserved1(int reserved1) {
		blockBuffer.putInt(RESERVED1_OFFSET, reserved1);
	}


	/**
	 * Main program for testing.
	 * 
	 * @param args ignored command line arguments.
	 */
	public static void main(String args[]) {
//      String infile  = "C:\\Documents and Settings\\heddle\\My Documents\\out.ev";
//      String outfile = "C:\\Documents and Settings\\heddle\\My Documents\\out_copy.ev";
        String infile  = "../../testdata/out.ev";
        String outfile = "../../testdata/out_copy.ev";
		int count = 0;

		try {
			//an EvioFile object is used for reading
			EvioFile inFile = new EvioFile(new File(infile));
			//an EvioWriter object is used for writing
			EventWriter eventWriter = new EventWriter(new File(outfile));
            //EventWriter eventWriter = new EventWriter(new File(outfile), 8192, true);

			EvioEvent event;
			while ((event = inFile.parseNextEvent()) != null) {				
//				System.out.println("EVENT LEN: " + event.getHeader().getLength());
				eventWriter.writeEvent(event);
				count++;
			}
			
			eventWriter.close();
		}
		catch (EvioException e) {
			e.printStackTrace();
		}
		catch (IOException e) {
			e.printStackTrace();
		}
		System.out.println("copied: " + count + " events.");
		
		//compare the two files
		EvioFile.compareEventFiles(new File(infile), new File(outfile));
	}


}
