package org.jlab.coda.jevio;

import java.util.List;
import java.util.Vector;

import javax.swing.tree.DefaultTreeModel;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

/**
 * An event is really just the outer, primary bank. That is, the first structure in an event (aka logical record, aka
 * buffer) must be a bank, probably a bank of banks.
 * 
 * The Event trivially extends bank, just so there can be a distinct <code>EvioEvent</code> class, for clarity.
 * 
 * @author heddle
 * 
 */
public class EvioEvent extends EvioBank {
		
	/**
	 * The XML record tag for an event.
	 */
	public static final String ELEMENT_NAME = "event";

	/**
	 * As the event is parsed, it is parsed into a tree.
	 */
	protected DefaultTreeModel treeModel;
	
	/**
	 * This is not the "num" field from the header, but rather a number [1..] indicating
	 * which event this was in the event file from which it was read. 
	 */
	private int eventNumber;


    /**
	 * Explicit null constructor for evio event.
	 */
	public EvioEvent() {
	}
	
    /**
     * Constructor using a provided BankHeader.
     *
     * @param bankHeader the header to use.
     * @see BankHeader
     */
    public EvioEvent(BankHeader bankHeader) {
        super(bankHeader);
    }

	/**
	 * This is a general constructor to use for an EvioEvent.
     *
	 * @param tag the tag for the event header (which is just a bank header).
	 * @param dataType the (enum) data type for the content of the bank.
	 * @param num sometimes, but not necessarily, an ordinal enumeration.
	 */
	public EvioEvent(int tag, DataType dataType, int num) {
		this(new BankHeader(tag, dataType, num));
	}

	/**
	 * This is a general constructor to use for an EvioEvent.
     *
	 * @param tag the tag for the event header (which is just a bank header).
	 * @param dataType the (int) data type for the content of the bank.
	 * @param num sometimes, but not necessarily, an ordinal enumeration.
	 */
	public EvioEvent(int tag, int dataType, int num) {
		this(new BankHeader(tag, dataType, num));
	}

	/**
	 * Create a string representation of the event.
	 * @return a string representation of the event.
	 */
	@Override
	public  String toString() {
		
		String description = getDescription();
		if (NameProvider.NO_NAME_STRING.equals(description)) {
			description = null;
		}
		
		String s = "<Event> " + " len (ints): " + header.length + " " + " tag: " + header.tag + " num: " + header.number;
		if (description != null) {
			s += " [" + description + "]";
		}
		return s;
	}
	
	/**
	 * Get the tree model representing this event. This would have been generated
	 * as the event was being parsed.
	 * @return the tree model representing this event.
	 */
	public DefaultTreeModel getTreeModel() {
		return treeModel;
	}
	
	/**
	 * Inserts a child structure into the event's JTree. This is called when the event is being parsed,
	 * and when an event is being created.
	 * @param child the child structure being added to the tree.
	 * @param parent the parent structure of the new child.
	 */
	public void insert(BaseStructure child, BaseStructure parent) {
		treeModel.insertNodeInto(child, parent, parent.getChildCount());
	}
	
	/**
	 * Visit all the structures in this event (including the event itself--which is considered its own descendant). 
	 * This is similar to listening to the event as it is being parsed, but is done to a complete (already) parsed event.
	 * @param listener an listener to notify as each structure is visited.
	 */
	public void vistAllStructures(IEvioListener listener) {
		visitAllDescendants(this, listener, null);
	}

	/**
	 * Visit all the structures in this event (including the event itself--which is considered its own descendant) in a depth first manner.
	 * @param listener an listener to notify as each structure is visited.
	 * @param filter an optional filter that must "accept" structures before they are passed to the listener. If <code>null</code>, all
	 * structures are passed. In this way, specific types of structures can be captured.
	 */
	public void vistAllStructures(IEvioListener listener, IEvioFilter filter) {
		visitAllDescendants(this, listener, filter);
	}

	/**
	 * Visit all the descendants of a given structure (which is considered a descendant of itself.)
	 * 
	 * @param structure the starting structure.
	 * @param listener an listener to notify as each structure is visited.
	 * @param filter an optional filter that must "accept" structures before they are passed to the listener. If
	 *            <code>null</code>, all structures are passed. In this way, specific types of structures can be
	 *            captured.
	 */
	private void visitAllDescendants(BaseStructure structure, IEvioListener listener, IEvioFilter filter) {
		
		if (listener != null) {
			boolean accept = true;
			if (filter != null) {
				accept = filter.accept(structure.getStructureType(), structure.header);
			}
			
			if (accept) {
			listener.gotStructure(this, structure);
			}
		}
		
		if (!(structure.isLeaf())) {
			for (BaseStructure child : structure.getChildren()) {
				visitAllDescendants(child, listener, filter);
			}
		}
	}
	
	/**
	 * Visit all the descendant structures, and collect those that pass a filter.
	 * @param filter the filter that must be passed. If <code>null</code>, this will return all the structures.
	 * @return a collection of all structures that are accepted by a filter.
	 */
	public List<BaseStructure> getMatchingStructures(IEvioFilter filter) {
		final Vector<BaseStructure> structures = new Vector<BaseStructure>(25, 10);
		
		IEvioListener listener = new IEvioListener() {
            @Override
            public void startEventParse(EvioEvent evioEvent) { }

            @Override
            public void endEventParse(EvioEvent evioEvent) { }
            
			@Override
			public void gotStructure(EvioEvent evioEvent, IEvioStructure structure) {
				structures.add((BaseStructure)structure);
			}
			
		};
		
		vistAllStructures(listener, filter);
		
		if (structures.size() == 0) {
			return null;
		}
		return structures;
	}
	
    /**
     * This returns a number [1..] indicating which event this was in the event file from which
     * it was read. It is not the "num" field from the header.
     * @return a number [1..] indicating which event this was in the event file from which
     * it was read.
     */
    public int getEventNumber() {
        return eventNumber;
    }

	/**
	 * This sets a number [1..] indicating which event this was in the event file from which
	 * it was read. It is not the "num" field from the header.
	 * @param eventNumber a number [1..] indicating which event this was in the event file from which
	 * it was read.
	 */
	public void setEventNumber(int eventNumber) {
		this.eventNumber = eventNumber;
	}

	/**
	 * Write this event structure out as an XML record.
	 * @param xmlWriter the writer used to write the events.
	 */
	@Override
	public void toXML(XMLStreamWriter xmlWriter) {

		try {
			int totalLen = header.length + 1;
			xmlWriter.writeComment(" Buffer " + getEventNumber() + " contains " + totalLen + " words (" + 4*totalLen + " bytes)");
			commonXMLStart(xmlWriter);
			xmlWriter.writeAttribute("format", "evio");
			xmlWriter.writeAttribute("count", "" + getEventNumber());
            if (DataType.getDataType(header.dataType).isStructure()) {
			    xmlWriter.writeAttribute("content", xmlContentAttributeName);
            }
			xmlWriter.writeAttribute("data_type", String.format("0x%x", header.dataType));
			xmlWriter.writeAttribute("length", "" + header.length);
			xmlWriter.writeAttribute("tag", "" + header.tag);
			xmlWriter.writeAttribute("num", "" + header.number);
			commonXMLClose(xmlWriter);
		}
		catch (XMLStreamException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Get the element name for the bank for writing to XML.
	 * @return the element name for the structure for writing to XML.
	 */
	@Override
	public String getXMLElementName() {
		return ELEMENT_NAME;
	}
	
}
