package org.jlab.coda.jevio;

import java.io.*;
import java.nio.*;
import java.nio.channels.FileChannel;

import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

/**
 * This is the a class of interest to the user. Create an <code>EvioFile</code> from a <code>File</code>
 * object corresponding to an event file, and from this class you can test the file for consistency and,
 * more importantly, you can call <code>parseNextEvent</code> to get new events and to stream the embedded
 * structures to an IEvioListener.
 * 
 * The streaming effect of parsing an event is that the parser will read the event and hand off structures,
 * such as banks, to any IEvioListeners. For those familiar with XML, the event is processed SAX-like.
 * It is up to the listener to decide what to do with the structures.
 * <p>
 * As an alternative to stream processing, after an event is parsed, the user can use the events treeModel
 * for access to the structures. For those familiar with XML, the event is processed DOM-like.
 * <p>
 * NOTE: Even though this class has a constructor that accepts an i/o mode, that is for backwards
 * compatibility only. An <code>EvioFile</code> is used for reading and parsing events only.
 * To write an event file, use an <code>EventWriter</code> object.
 *
 * @author heddle
 * 
 */
public class EvioFile {

	/**
	 * This <code>enum</code> denotes the status of a read. <br>
	 * SUCCESS indicates a successful read. <br>
	 * END_OF_FILE indicates that we cannot read because an END_OF_FILE has occurred. Technically this means that what
	 * ever we are trying to read is larger than the buffer's unread bytes.<br>
	 * EVIO_EXCEPTION indicates that an EvioException was thrown during a read, possibly due to out of range values,
	 * such as a negative start position.<br>
	 * UNKNOWN_ERROR indicates that an unrecoverable error has occurred.
	 */
	public static enum ReadStatus {
		SUCCESS, END_OF_FILE, EVIO_EXCEPTION, UNKNOWN_ERROR
	}

	/**
	 * This <code>enum</code> denotes the status of a write.<br>
	 * SUCCESS indicates a successful write. <br>
	 * CANNOT_OPEN_FILE indicates that we cannot write because the destination file cannot be opened.<br>
	 * EVIO_EXCEPTION indicates that an EvioException was thrown during a write.<br>
	 * UNKNOWN_ERROR indicates that an unrecoverable error has occurred.
	 */
	public static enum WriteStatus {
		SUCCESS, CANNOT_OPEN_FILE, EVIO_EXCEPTION, UNKNOWN_ERROR
	}

	/**
	 * Offset to get magic number from start of file.
	 */
	private static final int MAGIC_OFFSET = 28;
	
	/**
	 * The buffer representing a map of the input file.
	 */
	private MappedByteBuffer mappedByteBuffer;

    /**
     * Endianness of the data, either {@link ByteOrder#BIG_ENDIAN} or {@link ByteOrder#LITTLE_ENDIAN}.
     */
    private ByteOrder byteOrder;

	/**
	 * The current block header.
	 */
	private BlockHeader blockHeader = new BlockHeader();
	
	/**
	 * Root element tag for XML file
	 */
	private static final String ROOT_ELEMENT = "evio-data";

	/**
	 * Used to assign a transient number [1..n] to events as they are
	 */
	private int eventNumber = 0;

	/**
	 * This is the number of events in the file. It is not computed unless asked for, and if asked for it is computed
	 * and cached in this variable.
	 */
	private int eventCount = -1;

    /**
     * Absolute path of the underlying file.
     */
    private String path;

	
	/**
	 * Creates an event file for reading.
	 * 
	 * @param file the file that contains EVIO events.
	 * @throws IOException
	 */
	public EvioFile(File file) throws IOException {
		FileInputStream fileInputStream = new FileInputStream(file);
        path = new String(file.getAbsolutePath());
		FileChannel inputChannel = fileInputStream.getChannel();
		mapFile(inputChannel);
		inputChannel.close(); // this object is no longer needed
        // set buffer to beginning
        mappedByteBuffer.position(0);
    }

	/**
	 * Creates an event file.
	 * 
	 * @param path the full path to the file that contains events.
	 * For writing event files, use an <code>EventWriter</code> object.
	 * @see EventWriter
	 * @throws IOException
	 */
	public EvioFile(String path) throws IOException {
		this(new File(path));
	}

    /**
     * Get the path to the file.
     * @return path to the file
     */
    public String getPath() {
        return path;
    }

    /**
     * Get the number of events remaining in the file.
     *
     * @return number of events remaining in the file
     * @throws EvioException if failed reading from coda v3 file
     */
    public int getNumEventsRemaining() throws EvioException {
        return getEventCount() - eventNumber;
    }

	/**
	 * Maps the file into memory. The data are not actually loaded in memory-- subsequent reads will read
	 * random-access-like from the file.
	 * 
	 * @param inputChannel the input channel.
	 */
	private synchronized void mapFile(FileChannel inputChannel) throws IOException {
		long sz = inputChannel.size();
//System.out.println("mapFile: size = " + sz);
		mappedByteBuffer = inputChannel.map(FileChannel.MapMode.READ_ONLY, 0L, sz);
	}

	/**
	 * Obtain the file size using the memory mapped buffer's capacity, which should be the same.
	 * 
	 * @return the file size in bytes--actually the mapped memory size.
	 */
	public int fileSize() {
		return mappedByteBuffer.capacity();
	}

	/**
	 * Get the memory mapped buffer corresponding to the event file.
	 * 
	 * @return the memory mapped buffer corresponding to the event file.
	 */
	public MappedByteBuffer getMappedByteBuffer() {
		return mappedByteBuffer;
	}

	/**
	 * Reads the block (physical record) header. Assumes the mapped buffer is positioned
     * at the start of the next block header (physical record.)<br>
     *
	 * A Bank header is 8, 32-bit ints. The first int is the size of the block in ints
     * (not counting the length itself, i.e., the number of ints to follow).
	 * 
	 * Most users should have no need for this method, since most applications do care about the block
     * (physical record) header.
	 */
	protected synchronized ReadStatus nextBlockHeader() {
		// have enough remaining?
		if (mappedByteBuffer.remaining() < 32) {
			mappedByteBuffer.clear();
			return ReadStatus.END_OF_FILE;
		}

		// set the byte order to match the file's ordering
		try {
			// Check the magic number for endianness. This requires
			// peeking ahead 7 ints or 24 bytes. Set the endianness
            // once we figure out what it is (buffer defaults to big endian).
            byteOrder = mappedByteBuffer.order();
			int magicNumber = mappedByteBuffer.getInt(MAGIC_OFFSET);
//System.out.println("nextBlockHeader: actual magic number = " + Integer.toHexString(BlockHeader.MAGIC_NUMBER));
//System.out.println("nextBlockHeader: read magic number as " + Integer.toHexString(magicNumber));

			if (magicNumber != BlockHeader.MAGIC_NUMBER) {
                // Originally checked ByteOrder.nativeOrder() which is NOT what you want -  timmer
				if (byteOrder == ByteOrder.BIG_ENDIAN) {
                    byteOrder = ByteOrder.LITTLE_ENDIAN;
//System.out.println("nextBlockHeader: reset byte order to " + ByteOrder.LITTLE_ENDIAN);
				}
				else {
                    byteOrder = ByteOrder.BIG_ENDIAN;
//System.out.println("nextBlockHeader: reset byte order to " + ByteOrder.BIG_ENDIAN);
				}
                mappedByteBuffer.order(byteOrder);
			}
			else {
//System.out.println("nextBlockHeader: byte order = " + byteOrder);
			}

			try {
				// cache the starting position
				blockHeader.setBufferStartingPosition(mappedByteBuffer.position());

				// Now read the header data. It's a little trickier now that there are
                // several versions to be concerned about. Read all quantities into
                // temp variables, set the version FIRST, then set the other quantities last
                // as they may depend on the version number.
                int[] header = new int[8];
                for (int i=0; i<8; i++) {
                    header[i] = mappedByteBuffer.getInt();
                }

                int version = header[5] & 0xff;
                if (version != 1 && version != 2 && version != 3) {
                    // we got the wrong evio version here, throw an exception
                    throw new EvioException("Reading file of unsupported evio version (" + version + ")");
                }

                blockHeader.setVersion(version);
				blockHeader.setSize(header[0]);
				blockHeader.setNumber(header[1]);
				blockHeader.setHeaderLength(header[2]);
				blockHeader.setStart(header[3]);
				blockHeader.setEnd(header[4]);
				blockHeader.setReserved1(header[6]); // also event count in version 4
				blockHeader.setMagicNumber(header[7]);
			}
			catch (EvioException e) {
				e.printStackTrace();
				return ReadStatus.EVIO_EXCEPTION;
			}
		}
		catch (BufferUnderflowException a) {
System.err.println("ERROR endOfBuffer " + a);
			mappedByteBuffer.clear();
			return ReadStatus.UNKNOWN_ERROR;
		}

		return ReadStatus.SUCCESS;
	}


   /**
     * Get the next event in the file. As useful as this sounds, most applications will probably call
     * {@link #parseNextEvent() parseNextEvent} instead, since it combines combines getting the next
     * event with parsing the next event.<p>
     *
     * @return the next event in the file. On error it throws an EvioException. On end of file, it returns
     *         <code>null</code>.
     * @throws EvioException
     */
    public synchronized EvioEvent nextEvent() throws EvioException {
        EvioEvent event = new EvioEvent();
        BaseStructureHeader header = event.getHeader();

        // Are we at the top of the file?
        if (mappedByteBuffer.position() == 0) {
            ReadStatus status = nextBlockHeader();
            if (status != ReadStatus.SUCCESS) {
                throw new EvioException("Failed reading block header in nextEvent.");
            }
//System.out.println("nextEvent: BLOCK HEADER :\n" + blockHeader.toString());
        }

        // How many bytes remain in this block?
        // Must see if we have to deal with crossing physical record boundaries.
        int bytesRemaining = blockBytesRemaining();
        if (bytesRemaining < 0) {
            throw new EvioException("Number of block bytes remaining is negative.");
        }

        // Are we exactly at the end of the block (physical record)?
        if (bytesRemaining == 0) {
            ReadStatus status = nextBlockHeader();
            if (status == ReadStatus.SUCCESS) {
//System.out.println("nextEvent: BLOCK HEADER:\n" + blockHeader.toString());
                return nextEvent();
            }
            else if (status == ReadStatus.END_OF_FILE) {
                return null;
            }
            else {
                throw new EvioException("Failed reading block header in nextEvent.");
            }
        }
        // Or have we already read in the last event?
        else if (blockHeader.getBufferEndingPosition() == mappedByteBuffer.position()) {
            return null;
        }

        // No matter what, we can get the length of the next event.
        // A non positive length indicates eof;
        int length = mappedByteBuffer.getInt();
        if (length < 1) {
            return null;
        }

        header.setLength(length);
        bytesRemaining -= 4; // just read in 4 bytes

       // If we were unlucky, after reading the length there are no bytes remaining in this bank.
       if (bytesRemaining == 0) {
           ReadStatus status = nextBlockHeader();
           if (status == ReadStatus.END_OF_FILE) {
               return null;
           }
           else if (status != ReadStatus.SUCCESS) {
               throw new EvioException("Failed reading block header in nextEvent.");
           }
           bytesRemaining = blockBytesRemaining();
       }

        // Now should be good to go, except data may cross block boundary
        // in any case, should be able to read the rest of the header.
        if (byteOrder == ByteOrder.BIG_ENDIAN) {
            // interested in bit pattern, not negative numbers
            header.setTag(ByteDataTransformer.shortBitsToInt(mappedByteBuffer.getShort()));
            header.setDataType(ByteDataTransformer.byteBitsToInt(mappedByteBuffer.get()));
            header.setNumber(ByteDataTransformer.byteBitsToInt(mappedByteBuffer.get()));
        }
        else {
            header.setNumber(ByteDataTransformer.byteBitsToInt(mappedByteBuffer.get()));
            header.setDataType(ByteDataTransformer.byteBitsToInt(mappedByteBuffer.get()));
            header.setTag(ByteDataTransformer.shortBitsToInt(mappedByteBuffer.getShort()));
        }
        bytesRemaining -= 4; // just read in 4 bytes

        // get the raw data
        int eventDataSizeInts = header.getLength() - 1;
        int eventDataSizeBytes = 4 * eventDataSizeInts;

        try {
            byte bytes[] = new byte[eventDataSizeBytes];

            int bytesToGo = eventDataSizeBytes;
            int offset = 0;

            // be in while loop if have to cross block boundary[ies].
            while (bytesToGo > bytesRemaining) {
                mappedByteBuffer.get(bytes, offset, bytesRemaining);

                ReadStatus status = nextBlockHeader();
                if (status == ReadStatus.END_OF_FILE) {
                    return null;
                }
                else if (status != ReadStatus.SUCCESS) {
                    throw new EvioException("Failed reading block header after crossing boundary in nextEvent.");
                }
//System.out.println("nextEvent: BLOCK HEADER:\n" + blockHeader.toString());
                bytesToGo -= bytesRemaining;
                offset += bytesRemaining;
                bytesRemaining = blockBytesRemaining();
            }

            // last (perhaps only) read.
            mappedByteBuffer.get(bytes, offset, bytesToGo);
            event.setRawBytes(bytes);
            event.setByteOrder(byteOrder); // add this to track endianness, timmer
            event.setEventNumber(++eventNumber);
            return event;
        }
        catch (OutOfMemoryError ome) {
            System.err.println("Out Of Memory\n" +
                    "eventDataSizeBytes = " + eventDataSizeBytes + "\n" +
                            "bytes Remaining = " + bytesRemaining + "\n" +
                            "event Count: " + eventCount);
            return null;
        }
        catch (Exception e) {
            return null;
        }
    }

	/**
	 * This is the workhorse method. It retrieves the next event from the file, and then parses is SAX-like. It will
	 * drill down and uncover all structures (banks, segments, and tagsegments) and notify any interested listeners.
	 * 
	 * @return the event that was parsed. On error it throws an EvioException. On end of file, it returns
	 *         <code>null</code>.
	 * @throws EvioException
	 */
	public synchronized EvioEvent parseNextEvent() throws EvioException {
		EvioEvent event = nextEvent();
		if (event != null) {
			parseEvent(event);
		}
		return event;
	}

	/**
	 * This will parse an event, SAX-like. It will drill down and uncover all structures (banks, segments, and
	 * tagsegments) and notify any interested listeners.
	 * 
	 * As useful as this sounds, most applications will probably call {@link #parseNextEvent() parseNextEvent} instead,
	 * since it combines combines getting the next event with parsing the next event .
	 * 
	 * @param evioEvent the event to parse.
	 * @throws EvioException
	 */
	public synchronized void parseEvent(EvioEvent evioEvent) throws EvioException {
		EventParser.getInstance().parseEvent(evioEvent);
	}

	/**
	 * Get the number of bytes remaining in the current block (physical record). This is used for pathology checks like
	 * crossing the block boundary.
	 * 
	 * @return the number of bytes remaining in the current block (physical record).
	 */
	private int blockBytesRemaining() {
		try {
			return blockHeader.bytesRemaining(mappedByteBuffer.position());
		}
		catch (EvioException e) {
			e.printStackTrace();
			return -1;
		}
	}

	/**
	 * The equivalent of rewinding the file. What it actually does is set the position of the mapped memory buffer back
	 * to 0. This method, along with the two <code>position</code> and the <code>close</code> method, allows
	 * applications to treat this in a normal random access file manner.
	 */
	public void rewind() {
		mappedByteBuffer.position(0);
		eventNumber = 0;
	}

	/**
	 * This is equivalent to obtaining the current position in the file. What it actually does is return the position of
	 * the mapped memory buffer. This method, along with the <code>rewind</code>, <code>position(int)</code> and the
	 * <code>close</code> method, allows applications to treat this in a normal random access file manner.
	 * 
	 * @return the position of the file.
	 */
	public int position() {
		return mappedByteBuffer.position();
	}

	/**
	 * This is equivalent to setting the current position in the file. What it actually does is set the position of the
	 * mapped memory buffer. This method, along with the <code>rewind</code>, <code>position()</code> and the
	 * <code>close</code> method, allows applications to treat this in a normal random access file manner.
	 * 
	 * @param position the new position of the file.
	 */
	public void position(int position) {
		mappedByteBuffer.position(position);
	}

	/**
	 * This is equivalent to closing the file. What it actually does is clear all data from the mapped memory buffer,
	 * and sets its position to 0. This method, along with the <code>rewind</code> and the two <code>position()</code>
	 * methods, allows applications to treat this in a normal random access file manner.
	 */
	public void close() {
		mappedByteBuffer.clear();
	}

	/**
	 * This returns the current (active) block (physical record) header. Since most users have no interest in physical
	 * records, this method should not be used often. Mostly it is used by the test programs in the
	 * <code>EvioFileTest</code> class.
	 * 
	 * @return the current block header.
	 */
	public BlockHeader getCurrentBlockHeader() {
		return blockHeader;
	}

	/**
	 * Go to a specific event in the file. The events are numbered 1..N. This number is transient--it is not part of the
	 * event as stored in the evio file.
	 * 
	 * @param evNumber the event number in a 1..N counting sense, from the start of the file.
	 */
	public EvioEvent gotoEventNumber(int evNumber) {

		if (evNumber < 1) {
			return null;
		}

		// rewind
		rewind();
		EvioEvent event;

		try {
			// get the first evNumber - 1 events without parsing
			for (int i = 1; i < evNumber; i++) {
				event = nextEvent();
				if (event == null) {
					throw new EvioException("Asked to go to event: " + evNumber + ", which is beyond the end of file");
				}
			}
			// get one more event, the evNumber'th event, with parsing
			return parseNextEvent();
		}
		catch (EvioException e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Rewrite the entire file to XML.
	 * 
	 * @param path the path to the XML file.
	 * @return the status of the write.
	 */
	public WriteStatus toXMLFile(String path) {
		return toXMLFile(path, null);
	}

	/**
	 * Rewrite the entire file to XML.
	 * 
	 * @param path the path to the XML file.
	 * @param progressListener and optional progress listener, can be <code>null</code>.
	 * @return the status of the write.
	 * @see IEvioProgressListener
	 */
	public WriteStatus toXMLFile(String path, IEvioProgressListener progressListener) {
		FileOutputStream fos;

		try {
			fos = new FileOutputStream(path);
		}
		catch (FileNotFoundException e) {
			e.printStackTrace();
			return WriteStatus.CANNOT_OPEN_FILE;
		}

		XMLStreamWriter xmlWriter = null;
		try {
			xmlWriter = XMLOutputFactory.newInstance().createXMLStreamWriter(fos);
			xmlWriter.writeStartDocument();
			xmlWriter.writeCharacters("\n");
			xmlWriter.writeComment("Event source file: " + path);

			// start the root element
			xmlWriter.writeCharacters("\n");
			xmlWriter.writeStartElement(ROOT_ELEMENT);
			xmlWriter.writeAttribute("numevents", "" + getEventCount());

			// now loop through the events
			// rewind
			rewind();
			EvioEvent event;
			try {
				while ((event = parseNextEvent()) != null) {
					event.toXML(xmlWriter);
					// anybody interested in progress?
					if (progressListener != null) {
						progressListener.completed(event.getEventNumber(), getEventCount());
					}
				}
			}
			catch (EvioException e) {
				e.printStackTrace();
				return WriteStatus.UNKNOWN_ERROR;
			}

			// done. Close root element, end the document, and flush.
			xmlWriter.writeCharacters("\n");
			xmlWriter.writeEndElement();
			xmlWriter.writeEndDocument();
			xmlWriter.flush();
			xmlWriter.close();

			try {
				fos.close();
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		catch (XMLStreamException e) {
			e.printStackTrace();
			return WriteStatus.UNKNOWN_ERROR;
		}
        catch (FactoryConfigurationError e) {
            return WriteStatus.UNKNOWN_ERROR;
        }
        catch (EvioException e) {
            return WriteStatus.EVIO_EXCEPTION;
        }

		// TODO something better than just rewind again?
		rewind();
		System.out.println("XML write was successful");
		return WriteStatus.SUCCESS;
	}

    /**
     * This is the number of events in the file. It is not computed unless asked for, and if asked for it is computed
     * and cached.
     *
     * @return the number of events in the file.
     * @throws EvioException
     */
    public int getEventCount() throws EvioException {

        if (eventCount < 0) {
            rewind();
            eventCount = 0;

            while ((nextEvent()) != null) {
                eventCount++;
            }

            rewind();
        }

        return eventCount;
    }

	/**
	 * Method used for diagnostics. It compares two event files. It checks the following (in order):<br>
	 * <ol>
	 * <li> That neither file is null.
	 * <li> That both files exist.
	 * <li> That neither file is a directory.
	 * <li> That both files can be read.
	 * <li> That both files are the same size.
	 * <li> That both files contain the same number of events.
	 * <li> Finally, that they are the same in a byte-by-byte comparison.
	 * </ol> 
	 * NOTE: Two files with the same events but different physical record size will be reported as different. They will fail the same size test.
	 * @return <code>true</code> if the files are, byte-by-byte, identical.
	 */
	public static boolean compareEventFiles(File evFile1, File evFile2) {
		
		if ((evFile1 == null) || (evFile2 == null)) {
			System.out.println("In compareEventFiles, one or both files are null.");
			return false;
		}
		
		if (!evFile1.exists() || !evFile2.exists()) {
			System.out.println("In compareEventFiles, one or both files do not exist.");
			return false;
		}
		
		if (evFile1.isDirectory() || evFile2.isDirectory()) {
			System.out.println("In compareEventFiles, one or both files is a directory.");
			return false;
		}
		
		if (!evFile1.canRead() || !evFile2.canRead()) {
			System.out.println("In compareEventFiles, one or both files cannot be read.");
			return false;
		}
		
		String name1 = evFile1.getName();
		String name2 = evFile2.getName();
		
		long size1 = evFile1.length();
		long size2 = evFile2.length();
		
		if (size1 == size2) {
			System.out.println(name1 + " and " + name2 + " have the same length: " + size1);
		}
		else {
			System.out.println(name1 + " and " + name2 + " have the different lengths.");
			System.out.println(name1 + ": " + size1);
			System.out.println(name2 + ": " + size2);
			return false;
		}
		
		try {
			EvioFile evioFile1 = new EvioFile(evFile1);
			EvioFile evioFile2 = new EvioFile(evFile2);
			int evCount1 = evioFile1.getEventCount();
			int evCount2 = evioFile2.getEventCount();
			if (evCount1 == evCount2) {
				System.out.println(name1 + " and " + name2 + " have the same #events: " + evCount1);
			}
			else {
				System.out.println(name1 + " and " + name2 + " have the different #events.");
				System.out.println(name1 + ": " + evCount1);
				System.out.println(name2 + ": " + evCount2);
				return false;
			}
		}
        catch (EvioException e) {
            e.printStackTrace();
            return false;
        }
        catch (IOException e) {
            e.printStackTrace();
            return false;
        }


		System.out.print("Byte by byte comparison...");
		System.out.flush();
		
		int onetenth = (int)(1 + size1/10);
		
		//now a byte-by-byte comparison
		try {
			FileInputStream fis1 = new FileInputStream(evFile1);
			FileInputStream fis2 = new FileInputStream(evFile1);
			
			for (int i = 0; i < size1; i++) {
				try {
					int byte1 = fis1.read();
					int byte2 = fis2.read();
					
					if (byte1 != byte2) {
						System.out.println(name1 + " and " + name2 + " different at byte offset: " + i);
						return false;
					}
					
					if ((i % onetenth) == 0) {
						System.out.print(".");
						System.out.flush();
					}
				}
				catch (IOException e) {
					e.printStackTrace();
					return false;
				}
			}
			
			System.out.println("");
			
			try {
				fis1.close();
				fis2.close();
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		System.out.println("files " + name1 + " and " + evFile2.getPath() + " are identical.");
		return true;
	}


    /**
      * Array of ints, sequential, 1..size, for test purposes.
      * @param size the size of the array.
      * @return the fake int array.
      */
     private static int[] fakeIntArray(int size) {
         int[] array = new int[size];
         for (int i = 0; i < size; i++) {
             array[i] = 0x040302010;
         }
         return array;
     }

     /** For testing only */
     public static void main(String args[]) {
//         //create an event writer to write out the test events.
         String fileName  = "../../testdata/littleOut.ev";
//         EventWriter eventWriter = null;
//         File file = new File(fileName);
//
//         //count the events we make for testing
//         int eventNumber = 1;
//
//         //use a tag of 11 for events--for no particular reason
//         int tag = 11;
//
//         try {
//             eventWriter = new EventWriter(file);
//             //first event-- a trivial event containing an array of ints.
//             EventBuilder eventBuilder = new EventBuilder(tag, DataType.INT32, eventNumber++);
//             EvioEvent event1 = eventBuilder.getEvent();
//             //should end up with int array 1..25,1..10
//             eventBuilder.appendIntData(event1, fakeIntArray(10));
//             eventWriter.writeEvent(event1);
//         }
//         catch (EvioException e) {
//             e.printStackTrace();
//         }
//
//         // all done writing
//         eventWriter.close();


//         byte[] b = new byte[9];
//         b[0] = 67;
//         b[1] = 79;
//         b[2] = 0;
//         b[3] = 68;
//         b[4] = 65;
//         b[5] = 28;
//         b[6] = 29;
//         b[7] = 30;
//         b[8] = 31;
//
//         try {
//             String s = new String(b, "US-ASCII");
//             System.out.println("My string = " + s + ", length of string = " + s.length());
//
//             byte[] bs = s.getBytes("US-ASCII");
//             for (byte bb : bs){
//                 System.out.println("b = " + bb);
//             }
//
//             if (s.equals("CODA")) {
//                 System.out.println("CO0DA = CODA");
//             }
//             else {
//                 System.out.println("CO0DA != CODA");
//             }
//
//             StringBuffer stringData = new StringBuffer(s);
//             System.out.println("Buffer size = " + stringData.length());
//
//         }
//         catch (UnsupportedEncodingException e) {
//             e.printStackTrace();
//         }
//         System.exit(0);

         // Write Big Endian file (11 ints -> 8 block header ints, 2 bank header ints, 1 data int)
         ByteBuffer buf = ByteBuffer.allocate(4*8192);

         // block header is first
         buf.putInt(8192); // full block size
         buf.putInt(1);  // block number
         buf.putInt(8);  // header length
         buf.putInt(8);  // start of next event
         buf.putInt(60); // end of last event
         buf.putInt(2);  // version number
         buf.putInt(0);  // reserved
         buf.putInt(0xc0da0100);  // magic number

         //------------------------------
         // From here on we have an event
         //------------------------------

         // bank of banks next
         buf.putInt(51); // bank length
         //use a tag of 11 for events--for no particular reason
         int tag  = 11;
         int num  = 3;
         int type = 0x10 /* bank of banks */;
         buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

            // bank of banks next
            buf.putInt(18); // bank length
            tag  = 20;
            num  = 2;
            type = 0x10 /* bank of banks */;
            buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

            // bank of ints next
            buf.putInt(16); // bank length
            tag  = 21;
            num  = 1;
            type = 0xb /* bank of int32s */;
            buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
         buf.putInt(0x1);
         buf.putInt(0x2);
         buf.putInt(0x3);
         buf.putInt(0x4);
         buf.putInt(0x5);

         buf.putInt(0x10);
         buf.putInt(0x20);
         buf.putInt(0x30);
         buf.putInt(0x40);
         buf.putInt(0x50);

         buf.putInt(0x100);
         buf.putInt(0x200);
         buf.putInt(0x300);
         buf.putInt(0x400);
         buf.putInt(0x500);

            //------------------------

            // bank of segments next
            buf.putInt(3); // bank length
            tag  = 22;
            num  = 2;
            type = 0xd /* bank of segments */;
            buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

            // segment of ints next
            int len = 1; // seg length
            tag  = 23;
            type = 0xb /* segment of ints */;
            buf.putInt( (len & 0xffff) | ((type << 16) & 0xff0000) | (tag << 24) ); // segment header word
            buf.putInt(0x06050403);

            //------------------------

            // bank of tagsegments next
            buf.putInt(3); // bank length
            tag  = 24;
            num  = 2;
            type = 0xc /* bank of tagsegments */;
            buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

            // segment of ints next
            len  = 1; // tagseg length
            tag  = 25;
            type = 0xb /* tagsegment of ints */;
            buf.putInt( len & 0xffff | ((type << 16) & 0xf0000) | (tag << 20) ); // segment header word
            buf.putInt(0x07060504);


         // bank of ints next
         buf.putInt(2); // bank length
         tag  = 12;
         num  = 2;
         type = 0x1 /* bank of uint32s */;
         buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
         buf.putInt(0x04030201);

         // bank of shorts next
         buf.putInt(4); // bank length
         tag  = 13;
         num  = 2;
         type = 0x4 /* bank of shorts */;
         buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
         buf.putShort((short)0x0403);
         buf.putShort((short)0x0201);
         buf.putShort((short)0x0001);
         buf.putShort((short)0x0002);
         buf.putShort((short)0x0003);
         buf.putShort((short)0x0004);

         // bank of longs next
         buf.putInt(3); // bank length
         tag  = 14;
         num  = 2;
         type = 0x9 /* bank of longs */;
         buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
         buf.putLong(0x1234123412341234L);

         // bank of floats next
         buf.putInt(2); // bank length
         tag  = 15;
         num  = 2;
         type = 0x2 /* bank of floats */;
         buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
         buf.putFloat(1.23e-4F);

         // bank of doubles next
         buf.putInt(3); // bank length
         //use a tag of 11 for events--for no particular reason
         tag  = 16;
         num  = 2;
         type = 0x8 /* bank of doubles */;
         buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
         buf.putDouble(-5.6789e-300);

         // bank of strings
         buf.putInt(3); // bank length
         //use a tag of 11 for events--for no particular reason
         tag  = 17;
         num  = 2;
         type = 0x3 /* bank of strings */;
         buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
         byte[] b = new byte[8];
         b[0] = 67;
         b[1] = 79;
         b[2] = 68;
         b[3] = 65;
         b[4] = 51;
         b[5] = 15;
         buf.put(b);

         // now, write this into a file
         String fName  = "../../testdata/BigEndianOut.ev";
         File file2 = new File(fName);
         try {
             FileOutputStream fos = new FileOutputStream(file2);
             FileChannel channel = fos.getChannel();
             buf.rewind();
System.out.println("read ev file: " + fName + " size: " + file2.length());
             channel.write(buf);
         }
         catch (IOException e) {
             e.printStackTrace();
         }

         // parse bytes
         class myListener implements IEvioListener {

             public void startEventParse(EvioEvent evioEvent) { }

             public void endEventParse(EvioEvent evioEvent) { }

             public void gotStructure(EvioEvent evioEvent, IEvioStructure structure) {

                 BaseStructureHeader header = structure.getHeader();

                 System.out.println("------------------");
                 System.out.println("" + structure);

                 switch (header.getDataTypeEnum()) {
                     case FLOAT32:
                         System.out.println("        FLOAT VALS");
                         float floatdata[] = structure.getFloatData();
                         for (float f : floatdata) {
                             System.out.println("         " + f);
                         }
                         break;

                     case DOUBLE64:
                         System.out.println("        DOUBLE VALS");
                         double doubledata[] = structure.getDoubleData();
                         for (double d : doubledata) {
                             System.out.println("         " + d);
                         }
                         break;

                     case SHORT16:
                         System.out.println("        SHORT VALS");
                         short shortdata[] = structure.getShortData();
                         for (short i : shortdata) {
                             System.out.println("        0x" + Integer.toHexString(i));
                         }
                         break;

                     case INT32:
                     case UINT32:
                         System.out.println("        INT VALS");
                         int intdata[] = structure.getIntData();
                         for (int i : intdata) {
                             System.out.println("        0x" + Integer.toHexString(i));
                         }
                         break;

                     case LONG64:
                         System.out.println("        LONG VALS");
                         long longdata[] = structure.getLongData();
                         for (long i : longdata) {
                             System.out.println("        0x" + Long.toHexString(i));
                         }
                         break;

                     case CHAR8:
                     case UCHAR8:
                         System.out.println("        BYTE VALS");
                         byte bytedata[] = structure.getByteData();
                         for (byte i : bytedata) {
                             System.out.println("         " + i);
                         }
                         break;

                     case CHARSTAR8:
                         System.out.println("        STRING VALS");
                         String stringdata[] = structure.getStringData();
                         for (String str : stringdata) {
                             System.out.println("         " + str);
                         }
                         break;
                 }
             }
         }

         // jump past the block header
         buf.position(4*8);
         ByteBuffer buf2 = buf.slice();
   
         ByteParser parser = new ByteParser();
         parser.addEvioListener(new myListener());
         buf2.rewind();
         EvioEvent event = null;
         try {
             event = parser.parseEvent(buf2);
             System.out.println("Event = " + event.toString());
         }
         catch (EvioException e) {
             e.printStackTrace();
         }

         System.out.println("\nSWITCH BUFFER TO LITTLE ENDIAN HERE\n");

         // swap buffer
         EvioEvent event2;
         ByteBuffer dest = ByteBuffer.allocate(4*54);
         ByteBuffer swapAgain = ByteBuffer.allocate(4*54);
         byte[] fromArray = null, fromArraySwapped = null;
         try {
             buf2.rewind();
             //buf2.limit(4*20);
             ByteDataTransformer.swap(buf2, dest);
             dest.rewind();
             ByteDataTransformer.swap(dest, swapAgain);
             System.out.println("new buffer byte order is " + dest.order());

             // get array from event
             fromArray = event.toArray();

             // get array from swapped event
             event2 = ByteDataTransformer.swap(event);
             fromArraySwapped = event2.toArray();
         }
         catch (EvioException e) {
             e.printStackTrace();
         }

         // compare double swapped with original
         buf2.rewind();
         swapAgain.rewind();
         boolean notSame = false;
         for (int i=0; i<4*36; i++) {
             if (buf2.get() != swapAgain.get()) {
                 notSame = true;
                 break;
             }
         }
         if (notSame) {
             System.out.println("Double swapped buffer is different!!!");
         }
         else {
             System.out.println("Double swapped buffer is SAME!!!");
         }

         buf2.rewind();
         ByteBuffer fromBuf = ByteBuffer.wrap(fromArray);
         notSame = false;
         for (int i=0; i<4*36; i++) {
             if (buf2.get() != fromBuf.get()) {
                 notSame = true;
                 break;
             }
         }
         if (notSame) {
             System.out.println("toArray array is different!!!");
         }
         else {
             System.out.println("toArray array is SAME!!!");
         }


         dest.rewind();
         ByteBuffer fromBufSwap = ByteBuffer.wrap(fromArraySwapped);
         notSame = false;
         for (int i=0; i<4*36; i++) {
             if (dest.get() != fromBufSwap.get()) {
                 notSame = true;
                 break;
             }
         }
         if (notSame) {
             System.out.println("toArray array swapped is different!!!");
         }
         else {
             System.out.println("toArray array swapped is SAME!!!");
         }

         dest.rewind();
         try {
             event = parser.parseEvent(dest);
             System.out.println("Event (dest)  = " + event.toString());
         }
         catch (EvioException e) {
             e.printStackTrace();
         }

         // now read it back in

         File fileIn = new File(fileName);
         System.out.println("read ev file: " + fileName + " size: " + fileIn.length());

         try {
             EvioFile evioFile = new EvioFile(fileName);
             MappedByteBuffer buffer = evioFile.getMappedByteBuffer();
             // print out event data in hex:
         }
         catch (IOException e) {
             e.printStackTrace();
         }
     }


}
