package org.jlab.coda.jevio;

import org.jlab.coda.jevio.EvioFile.ReadStatus;

/**
 * A set of static functions that test evio files. It also has some other diagnostic methods for getting counts.
 * 
 * @author heddle
 * 
 */
public class EvioFileTest {

	/**
	 * This enum is used for file testing. <br>
	 * PASS indicates that a given test passed. <br>
	 * FAIL indicates that a given test failed.
	 */
	public static enum TestResult {
		PASS, FAIL
	}

	/**
	 * Get a total count of the number of physical records.
	 * 
	 * @param evioFile the file to be processed.
	 * @return the total count of blocks (physical records.)
	 */
	public static int totalBlockCount(EvioFile evioFile) {
		evioFile.rewind();
		ReadStatus status = ReadStatus.SUCCESS;
		int count = 0;

		while (status == ReadStatus.SUCCESS) {
			status = evioFile.nextBlockHeader();
			if (status == ReadStatus.SUCCESS) {
				evioFile.position(evioFile.getCurrentBlockHeader().nextBufferStartingPosition());
				count++;
			}
		}

		System.out.println("total block count: " + count);

		evioFile.rewind();
		return count;
	} // totalBlockCount

	/**
	 * Tests whether we can look through the file and find all the block headers.
	 * 
	 * @param evioFile the file to be tested.
	 * @return the result of this test, either PASS or FAIL.
	 */
	public static TestResult readAllBlockHeadersTest(EvioFile evioFile) {
		evioFile.rewind();
		ReadStatus status = ReadStatus.SUCCESS;
		int blockCount = 0;

		while (status == ReadStatus.SUCCESS) {
			status = evioFile.nextBlockHeader();
			if (status == ReadStatus.SUCCESS) {
				evioFile.position(evioFile.getCurrentBlockHeader().nextBufferStartingPosition());
				blockCount++;
			}
		}

		TestResult result;

		// should have read to the end of the file
		if (status == ReadStatus.END_OF_FILE) {
			System.out.println("Total blocks read: " + blockCount);
			result = TestResult.PASS;
		}
		else {
			result = TestResult.FAIL;
		}

		System.out.println("readAllBlockHeadersTest: " + result);

		evioFile.rewind();
		return result;
	} // readAllBlockHeadersTest

	/**
	 * Tests whether we can look through the file read all the events.
	 * 
	 * @param evioFile the file to be tested.
	 * @return the result of this test, either <code>TestResult.PASS</code> or <code>TestResult.FAIL</code>.
	 */
	public static TestResult readAllEventsTest(EvioFile evioFile) {
		// store current file position
		int oldPosition = evioFile.position();

		evioFile.rewind();
		EvioEvent event;
		int count = 0;
		TestResult result = TestResult.PASS;

		try {
			while ((event = evioFile.nextEvent()) != null) {
				count++;
				BaseStructureHeader header = event.getHeader();

				System.out.println(count + ")  size: " + header.getLength() + " type: " + header.getDataTypeName()
						+ " \"" + event.getDescription() + "\"");
			}
		}
		catch (EvioException e) {
			e.printStackTrace();
			result = TestResult.FAIL;
		}

		System.out.println("readAllBlockHeadersTest: " + result);

		// restore file position
		evioFile.position(oldPosition);
		return result;
	} // readAllEventsTest

	/**
	 * Tests whether we can parse events from the file.
	 * 
	 * @param evioFile the file to be tested.
	 * @param num the number to parse. Will try to parse this many (unless it runs out.) Use -1 to parse all events.
	 *            Note: if <code>num</code> is greater than the number of events in the file, it doesn't constitute an
	 *            error.
	 * @return the result of this test, either <code>TestResult.PASS</code> or <code>TestResult.FAIL</code>.
	 */
	public static TestResult parseEventsTest(EvioFile evioFile, int num) {
		// store current file position
		int oldPosition = evioFile.position();

		if (num < 0) {
			num = Integer.MAX_VALUE;
		}

		evioFile.rewind();
		EvioEvent event;
		int count = 0;
		TestResult result = TestResult.PASS;

		try {
			while ((count < num) && ((event = evioFile.nextEvent()) != null)) {
				evioFile.parseEvent(event);
				count++;
			}
		}
		catch (EvioException e) {
			e.printStackTrace();
			result = TestResult.FAIL;
		}

		System.out.println("parseEventsTest parsed: " + count + " events");
		System.out.println("parseEventsTest result: " + result);

		// restore file position
		evioFile.position(oldPosition);
		return result;
	} // readAllEventsTest

}
