package org.jlab.coda.jevio;

import java.io.File;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.util.Collections;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * This was developed to read the xml dictionary that Maurizio uses for GEMC. It implements INameProvider, just like all
 * other dictionary readers.
 * 
 * @author heddle
 * 
 */
@SuppressWarnings("serial")
public class EvioXMLDictionary extends Vector<EvioDictionaryEntry> implements INameProvider {

	/**
	 * There is only one type of element. This is its stange name.
	 */
	private static String ENTRY = "xmldumpDictEntry";

	/**
	 * The "name" attribute name.
	 */
	private static String NAME = "name";

	/**
	 * The "tag" attribute name.
	 */
	private static String TAG = "tag";

	/**
	 * The "num" attribute name.
	 */
	private static String NUM = "num";

    /**
     * Create an EvioXMLDictionary. This is not a standard CODA disctionary, but a format developed for use by GEMC.
     *
     * @param file file containing xml.
     */
    public EvioXMLDictionary(File file) {
        this(getDomObject(file));
    }

    /**
     * Create an EvioXMLDictionary. This is not a standard CODA disctionary, but a format developed for use by GEMC.
     *
     * @param xmlString string containing xml.
     */
    public EvioXMLDictionary(String xmlString) {
        this(getDomObject(xmlString));
    }

    /**
	 * Create an EvioXMLDictionary. This is not a standard CODA disctionary, but a format developed for use by GEMC.
	 * 
	 * @param domDocument DOM object representing xml dictionary.
	 */
	public EvioXMLDictionary(Document domDocument) {
		super(100);

		// try to get the XML dom document
		if (domDocument != null) {
			NodeList list = domDocument.getElementsByTagName(ENTRY);

			if (list != null) {
				for (int index = 0; index < list.getLength(); index++) {
					Node node = list.item(index);
					if (node != null) {
						if (node.hasAttributes()) {
							String name = null;
							String tag = null;
							String num = null;

							NamedNodeMap map = node.getAttributes();

							// get the name
							Node nameNode = map.getNamedItem(NAME);
							if (nameNode != null) {
								name = nameNode.getNodeValue();
							}

							// get the tag
							Node tagNode = map.getNamedItem(TAG);
							if (tagNode != null) {
								tag = tagNode.getNodeValue();
							}

							// get the num
							Node numNode = map.getNamedItem(NUM);
							if (numNode != null) {
								num = numNode.getNodeValue();
							}
							
							add (new EvioDictionaryEntry(tag, num, name));
						} // end node has attributes
					} // end node not null
				} // end loop over list
			}
		}
		
		//sort
		Collections.sort(this);
		
//System.out.println(this);

	} // end Constructor

	
	/**
	 * Returns the pretty name of some evio structure. Typically this is involve
	 * the use of the "tag" and, if present, "num" fields. There may also be a hierarchical
	 * dependence. 
	 * 
	 * @param structure the structure to find the name of.
	 * @return a descriptive name, e.g., "Edep".
	 */
	@Override
	public String getName(BaseStructure structure) {
		for (EvioDictionaryEntry entry : this) {
			if (entry.match(structure)) {
				return entry.getDescription(); 
			}
		}
		return NameProvider.NO_NAME_STRING;
	}

	/**
	 * Return a dom object corresponding to an xml file.
	 * 
	 * @param file the XML File object
	 * @return the dom object (or <code>null</code>.)
	 */
	private static Document getDomObject(File file) {
		// get the factory
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		Document dom = null;

		try {

			// Using factory get an instance of document builder
			DocumentBuilder db = dbf.newDocumentBuilder();

			// parse using builder to get DOM representation of the XML file
			dom = db.parse(file);

		}
		catch (ParserConfigurationException e) {
			e.printStackTrace();
		}
		catch (SAXException e) {
			e.printStackTrace();
		}
		catch (IOException e) {
			e.printStackTrace();
		}

		return dom;
	}

    /**
     * Return a dom object corresponding to an xml string.
     *
     * @param xmlString XML string representing dictionary
     * @return the dom object (or <code>null</code>.)
     */
    private static Document getDomObject(String xmlString) {
        // get the factory
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        Document dom = null;

        try {

            // Using factory get an instance of document builder
            DocumentBuilder db = dbf.newDocumentBuilder();

            // parse using builder to get DOM representation of the XML string
            ByteArrayInputStream bais = new ByteArrayInputStream(xmlString.getBytes());
            dom = db.parse(bais);

        }
        catch (ParserConfigurationException e) {
            e.printStackTrace();
        }
        catch (SAXException e) {
            e.printStackTrace();
        }
        catch (IOException e) {
            e.printStackTrace();
        }

        return dom;
    }

	/**
	 * Get a string representation of the dictionary.
	 * @return a string representation of the dictionary.
	 */
	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer(1024);
		sb.append("-- Dictionary --\n");
		for (EvioDictionaryEntry entry : this) {
			sb.append(entry + "\n");
		}
		return sb.toString();
	}
}
