package org.jlab.coda.jevio;

import java.util.List;

/**
 * This is a set of convenient static methods used to find lists of structures within an event that
 * match certain criteria. For most part it uses the <code>List<BaseStructure>&nbsp;getMatchingStructures(IEvioFilter)</code> 
 * method on the provided <code>EvioEvent</code> object by constructing the appropriate filter.
 * @author heddle
 *
 */
public class StructureFinder {

	
	/**
	 * Collect all the structures in an event that pass a filter.
	 * @param event the event being queried.
	 * @param filter the filter that must be passed. If <code>null</code>, this will return all the structures.
	 * @return a collection of all structures that are accepted by a filter for the provided event.
	 */
	public static List<BaseStructure> getMatchingStructures(EvioEvent event, IEvioFilter filter) {
		if (event == null) {
			return null;
		}
		return event.getMatchingStructures(filter);
	}
	
	/**
	 * Collect all the banks in an event that match a provided tag and number in their header. Only Banks are returned,
	 * because only Banks have a number field.
	 * @param event the event being queried.
	 * @param tag the tag to match.
	 * @param number the number to match.
	 * @return a collection of all Banks that are accepted by a filter for the provided event.
	 */
	public static List<BaseStructure> getMatchingBanks(EvioEvent event, final int tag, final int number) {
		IEvioFilter filter = new IEvioFilter() {
			@Override
			public boolean accept(StructureType type, BaseStructureHeader header) {
				return (type == StructureType.BANK) && (tag == header.tag) && (number == header.number);
			}
		};
		return getMatchingStructures(event, filter);
	}	
	
	/**
	 * Collect all the structures in an event that match a provided tag in their header.
	 * @param event the event being queried.
	 * @param tag the tag to match.
	 * @return a collection of all structures that are accepted by a filter for the provided event.
	 */
	public static List<BaseStructure> getMatchingStructures(EvioEvent event, final int tag) {
		IEvioFilter filter = new IEvioFilter() {
			@Override
			public boolean accept(StructureType type, BaseStructureHeader header) {
				return (tag == header.tag);
			}
		};
		return getMatchingStructures(event, filter);
	}	

	/**
	 * Collect all the non-banks (i.e., Segments and TagSegments) in an event that match a provided tag in their header. No Banks are returned,
	 * @param event the event being queried.
	 * @param tag the tag to match.
	 * @return a collection of all non-bank structures that are accepted by a filter for the provided event.
	 */
	public static List<BaseStructure> getMatchingNonBanks(EvioEvent event, final int tag) {
		IEvioFilter filter = new IEvioFilter() {
			@Override
			public boolean accept(StructureType type, BaseStructureHeader header) {
				return (type != StructureType.BANK) && (tag == header.tag);
			}
		};
		return getMatchingStructures(event, filter);
	}	

}
