/*----------------------------------------------------------------------------*
 *  Copyright (c) 2009        Jefferson Science Associates,                   *
 *                            Thomas Jefferson National Accelerator Facility  *
 *                                                                            *
 *    This software was developed under a United States Government license    *
 *    described in the NOTICE file included as part of this distribution.     *
 *                                                                            *
 *    C. Timmer,  Oct 20, 2009, Jefferson Lab                                 *
 *                                                                            *
 *     Author: Carl Timmer                                                    *
 *             timmer@jlab.org                   Jefferson Lab, MS-12B3       *
 *             Phone: (757) 269-5130             12000 Jefferson Ave.         *
 *             Fax:   (757) 269-6248             Newport News, VA 23606       *
 *                                                                            *
 *----------------------------------------------------------------------------*/

package org.jlab.coda.jevio;


import org.jlab.coda.cMsg.*;

import javax.xml.stream.XMLStreamWriter;
import javax.xml.stream.XMLOutputFactory;
import java.nio.ByteBuffer;
import java.io.StringWriter;
import java.io.PrintWriter;


/**
 * This is an example class which creates a cMsg message producer of evio events.
 */
public class Test {

    private String  subject = "evio";
    private String  type = "event";
    private String  name = "evoiProducer";
    private String  description = "evio producer";
    private String  UDL = "cMsg://localhost/cMsg/myNameSpace";


    private int     delay, count = 50000;
    private boolean debug;


    /** Constructor. */
    Test(String[] args) {
        decodeCommandLine(args);
    }


    /**
     * Method to decode the command line used to start this application.
     * @param args command line arguments
     */
    private void decodeCommandLine(String[] args) {

        // loop over all args
        for (int i = 0; i < args.length; i++) {

            if (args[i].equalsIgnoreCase("-h")) {
                usage();
                System.exit(-1);
            }
            else if (args[i].equalsIgnoreCase("-n")) {
                name = args[i + 1];
                i++;
            }
            else if (args[i].equalsIgnoreCase("-d")) {
                description = args[i + 1];
                i++;
            }
            else if (args[i].equalsIgnoreCase("-u")) {
                UDL= args[i + 1];
                i++;
            }
            else if (args[i].equalsIgnoreCase("-s")) {
                subject = args[i + 1];
                i++;
            }
            else if (args[i].equalsIgnoreCase("-t")) {
                type = args[i + 1];
                i++;
            }
            else if (args[i].equalsIgnoreCase("-c")) {
                count = Integer.parseInt(args[i + 1]);
                if (count < 1)
                    System.exit(-1);
                i++;
            }
            else if (args[i].equalsIgnoreCase("-delay")) {
                delay = Integer.parseInt(args[i + 1]);
                i++;
            }
            else if (args[i].equalsIgnoreCase("-debug")) {
                debug = true;
            }
            else {
                usage();
                System.exit(-1);
            }
        }

        return;
    }


    /** Method to print out correct program command line usage. */
    private static void usage() {
        System.out.println("\nUsage:\n\n" +
            "   java Test\n" +
            "        [-n <name>]          set client name\n"+
            "        [-d <description>]   set description of client\n" +
            "        [-u <UDL>]           set UDL to connect to cMsg\n" +
            "        [-s <subject>]       set subject of sent messages\n" +
            "        [-t <type>]          set type of sent messages\n" +
            "        [-c <count>]         set # of messages to send before printing output\n" +
            "        [-delay <time>]      set time in millisec between sending of each message\n" +
            "        [-debug]             turn on printout\n" +
            "        [-h]                 print this help\n");
    }

    /**
     * Run as a stand-alone application.
     */
    public static void main(String[] args) {
        try {
            Test producer = new Test(args);
            producer.run();
        }
        catch (cMsgException e) {
            System.out.println(e.toString());
            System.exit(-1);
        }
    }


    /**
     * Method to convert a double to a string with a specified number of decimal places.
     *
     * @param d double to convert to a string
     * @param places number of decimal places
     * @return string representation of the double
     */
    private static String doubleToString(double d, int places) {
        if (places < 0) places = 0;

        double factor = Math.pow(10,places);
        String s = "" + (double) (Math.round(d * factor)) / factor;

        if (places == 0) {
            return s.substring(0, s.length()-2);
        }

        while (s.length() - s.indexOf(".") < places+1) {
            s += "0";
        }

        return s;
    }



    /**
      * This method is executed as a thread.
      */
     public void run() throws cMsgException {

        int i, j=0, k;
        i = j++;
        k = j++ % 1000;
        System.out.println("I =  " + i + ", k = " + k + ", j = " + j);
        if (true) return;

         if (debug) {
             System.out.println("Running jevio parsing level test");
         }

        try {
// traditional bank of banks
            EventBuilder eventBuilder = new EventBuilder(0, DataType.BANK, 0);
            EvioEvent event = eventBuilder.getEvent();

            // add a bank of ints to top bank of banks (must be level = 1 to parse these ints)
            EvioBank bank01 = new EvioBank(0x1, DataType.INT32, 1);
            int[] iarray = {777};
            eventBuilder.appendIntData(bank01, iarray);
            eventBuilder.addChild(event, bank01);

            // add a bank of segs - Level 1 (top level fully parsed, this level not)
            EvioBank bank1 = new EvioBank(0x1, DataType.SEGMENT, 1); // tag, type, num
            eventBuilder.addChild(event, bank1);

            // add a seg of banks - Level 2
            EvioSegment segment2 = new EvioSegment(0x2, DataType.BANK);
            eventBuilder.addChild(bank1, segment2);

            // add a bank of banks - Level 3
            EvioBank bank3 = new EvioBank(0x3, DataType.TAGSEGMENT, 3);
            eventBuilder.addChild(segment2, bank3);

            // add a bank of tagsegs - Level 4
            EvioTagSegment tagsegment4 = new EvioTagSegment(0x4, DataType.BANK);
            eventBuilder.addChild(bank3, tagsegment4);

            // add a tagsegs of banks - Level 5
            EvioBank bank5 = new EvioBank(0x5, DataType.BANK, 5);
            eventBuilder.addChild(tagsegment4, bank5);

            // add a bank of banks - Level 6
            EvioBank bank6 = new EvioBank(0x6, DataType.BANK, 6);
            eventBuilder.addChild(bank5, bank6);

            // add a bank of ints to last bank of banks (must be level = 7 to parse the ints)
            EvioBank bank61 = new EvioBank(0x61, DataType.INT32, 61);
            int[] iarray2 = {666};
            eventBuilder.appendIntData(bank61, iarray2);
            eventBuilder.addChild(bank6, bank61);

            // parse
            ByteBuffer bbuf = ByteBuffer.allocate(1000);

            bbuf.clear();
            event.write(bbuf);
            bbuf.flip();

            ByteParser parser = new ByteParser();
            EvioEvent event2Parse = parser.parseEvent(bbuf);

            StringWriter sw = new StringWriter(1000);
            XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance().createXMLStreamWriter(sw);
            event2Parse.toXML(xmlWriter);
            System.out.println("Parsed Event:\n" + sw.toString());


        }
        catch (Exception e) {
            e.printStackTrace();
        }

//                        // add a bank of segments
//                        EvioBank bank2 = new EvioBank(0x33, DataType.SEGMENT, 0);
//
//                        EvioSegment segment1 = new EvioSegment(0x34, DataType.SHORT16);
//                        short[] sarray = {4,5,6};
//                        eventBuilder.appendShortData(segment1, sarray);
//                        eventBuilder.addChild(bank2, segment1);
//                        eventBuilder.addChild(event, bank2);
//
//                        // add a bank of tag segments
//                        EvioBank bank3 = new EvioBank(0x45, DataType.TAGSEGMENT, 0);
//
//                        EvioTagSegment tagsegment1 = new EvioTagSegment(0x46, DataType.DOUBLE64);
//                        double[] darray = {7,8,9};
//                        eventBuilder.appendDoubleData(tagsegment1, darray);
//                        eventBuilder.addChild(bank3, tagsegment1);
//                        eventBuilder.addChild(event, bank3);
//



    }



    /**
     * This method is executed as a thread.
     */
    public void run2() throws cMsgException {

        if (debug) {
            System.out.println("Running cMsg producer sending to:\n" +
                                 "    subject = " + subject +
                               "\n    type    = " + type +
                               "\n    UDL     = " + UDL);
        }

        // connect to cMsg server
        cMsg coda = new cMsg(UDL, name, description);
        try {
            coda.connect();
        }
        catch (cMsgException e) {
            e.printStackTrace();
            return;
        }

        // create a message
        cMsgMessage msg = new cMsgMessage();
        msg.setSubject(subject);
        msg.setType(type);

        // add binary array containing evio data
        ByteBuffer buf = ByteBuffer.allocate(4*8192);

        // bank of banks next
        buf.putInt(51); // bank length
        //use a tag of 11 for events--for no particular reason
        int tag  = 11;
        int num  = 3;
        int type = 0x10 /* bank of banks */;
        buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

           // bank of banks next
           buf.putInt(18); // bank length
           tag  = 20;
           num  = 2;
           type = 0x10 /* bank of banks */;
           buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

           // bank of ints next
           buf.putInt(16); // bank length
           tag  = 21;
           num  = 1;
           type = 0xb /* bank of int32s */;
           buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

           buf.putInt(0x1);
           buf.putInt(0x2);
           buf.putInt(0x3);
           buf.putInt(0x4);
           buf.putInt(0x5);

           buf.putInt(0x10);
           buf.putInt(0x20);
           buf.putInt(0x30);
           buf.putInt(0x40);
           buf.putInt(0x50);

           buf.putInt(0x100);
           buf.putInt(0x200);
           buf.putInt(0x300);
           buf.putInt(0x400);
           buf.putInt(0x500);

           //------------------------

           // bank of segments next
           buf.putInt(3); // bank length
           tag  = 22;
           num  = 2;
           type = 0xd /* bank of segments */;
           buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

           // segment of ints next
           int len = 1; // seg length
           tag  = 23;
           type = 0xb /* segment of ints */;
           buf.putInt( (len & 0xffff) | ((type << 16) & 0xff0000) | (tag << 24) ); // segment header word
           buf.putInt(0x06050403);

           //------------------------

           // bank of tagsegments next
           buf.putInt(3); // bank length
           tag  = 24;
           num  = 2;
           type = 0xc /* bank of tagsegments */;
           buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

           // segment of ints next
           len  = 1; // tagseg length
           tag  = 25;
           type = 0xb /* tagsegment of ints */;
           buf.putInt( len & 0xffff | ((type << 16) & 0xf0000) | (tag << 20) ); // segment header word
           buf.putInt(0x07060504);


        // bank of ints next
        buf.putInt(2); // bank length
        tag  = 12;
        num  = 2;
        type = 0x1 /* bank of uint32s */;
        buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
        buf.putInt(0x04030201);

        // bank of shorts next
        buf.putInt(4); // bank length
        tag  = 13;
        num  = 2;
        type = 0x4 /* bank of shorts */;
        buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word

        buf.putShort((short)0x0403);
        buf.putShort((short)0x0201);
        buf.putShort((short)0x0001);
        buf.putShort((short)0x0002);
        buf.putShort((short)0x0003);
        buf.putShort((short)0x0004);

        // bank of longs next
        buf.putInt(3); // bank length
        tag  = 14;
        num  = 2;
        type = 0x9 /* bank of longs */;
        buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
        buf.putLong(0x1234123412341234L);

        // bank of floats next
        buf.putInt(2); // bank length
        tag  = 15;
        num  = 2;
        type = 0x2 /* bank of floats */;
        buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
        buf.putFloat(1.23e-4F);

        // bank of doubles next
        buf.putInt(3); // bank length
        //use a tag of 11 for events--for no particular reason
        tag  = 16;
        num  = 2;
        type = 0x8 /* bank of doubles */;
        buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
        buf.putDouble(-5.6789e-300);

        // bank of strings
        buf.putInt(3); // bank length
        //use a tag of 11 for events--for no particular reason
        tag  = 17;
        num  = 2;
        type = 0x3 /* bank of strings */;
        buf.putInt( (num & 0xff) | ((type << 8) & 0xff00) | (tag << 16) ); // second bank header word
        byte[] b = new byte[8];
        b[0] = 67;
        b[1] = 79;
        b[2] = 68;
        b[3] = 65;
        b[4] = 51;
        b[5] = 15;
        buf.put(b);
        buf.flip();

        msg.setByteArrayNoCopy(buf.array(), 0, 4*60);
        msg.setByteArrayEndian(cMsgConstants.endianBig);

        // Add payload dictionary
        String xml = "<?xml version=\"1.0\" ?>" +
"<!-- dictionary corresponding to evio event -->" +
"<dictionary>" +
"    <!-- dictionary entries -->" +
"    <xmldumpDictEntry name=\"cMsgEvent\"        tag=\"11\"   num=\"3\"/>" +
"    <xmldumpDictEntry name=\"cMsgBank1\"        tag=\"20\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgIntBank1\"     tag=\"21\"   num=\"2.1\"/>" +
"    <xmldumpDictEntry name=\"cMsgBank2\"        tag=\"22\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgIntSeg1\"      tag=\"23\"   />" +
"    <xmldumpDictEntry name=\"cMsgBank3\"        tag=\"24\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgIntTagSeg1\"   tag=\"25\"  />" +
"    <xmldumpDictEntry name=\"cMsgIntBank2\"     tag=\"12\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgShortBank1\"   tag=\"13\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgLongBank1\"    tag=\"14\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgFloatBank1\"   tag=\"15\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgDoubleBank1\"  tag=\"16\"   num=\"2\"/>" +
"    <xmldumpDictEntry name=\"cMsgStringBank1\"  tag=\"17\"   num=\"2\"/>" +
"</dictionary>";
        cMsgPayloadItem dictionaryItem = new cMsgPayloadItem("dictionary", xml);
        msg.addPayloadItem(dictionaryItem);


        // variables to track message rate
        double freq=0., freqAvg=0.;
        long t1, t2, deltaT, totalT=0, totalC=0;

        // Ignore the first N values found for freq in order
        // to get better avg statistics. Since the JIT compiler in java
        // takes some time to analyze & compile code, freq may initially be low.
        long ignore=0;

        while (true) {
            t1 = System.currentTimeMillis();
            for (int i = 0; i < count; i++) {
                coda.send(msg);
                coda.flush(0);

                // delay between messages sent
                if (delay != 0) {
                    try {Thread.sleep(delay);}
                    catch (InterruptedException e) {}
                }
            }
            t2 = System.currentTimeMillis();

            if (ignore == 0) {
                deltaT = t2 - t1; // millisec
                freq = (double) count / deltaT * 1000;
                totalT += deltaT;
                totalC += count;
                freqAvg = (double) totalC / totalT * 1000;

                if (debug) {
                    System.out.println(doubleToString(freq, 1) + " Hz, Avg = " +
                                       doubleToString(freqAvg, 1) + " Hz");
                }
            }
            else {
                ignore--;
            }
        }
    }

}